
/////////////////////////////////////////////////////////////
//                                                         //
//   AcedCompression 1.03                                  //
//                                                         //
//   ,        //
//   LZ+Huffman    .               //
//                                                         //
//   mailto: acedutils@yandex.ru                           //
//                                                         //
/////////////////////////////////////////////////////////////

unit AcedCompression;

interface

{  ,      ,
     . }

type
  TCompressionMode = (
    dcmNoCompression,       //      
    dcmFastest,             //  ,      
    dcmFast,                //      
    dcmNormal,              //      
    dcmMaximumCompression   //   
  );

{ G_Deflate   ,   SourceBytes,
   SourceLength .     Mode.  
     ,     . 
  ,       BeforeGap ,    -
   AfterGap .    ,   ,
     DestinationBytes.    
  (  BeforeGap  AfterGap)    . }

function G_Deflate(SourceBytes: Pointer; SourceLength: Integer;
  var DestinationBytes: Pointer; BeforeGap, AfterGap: Integer;
  Mode: TCompressionMode = dcmFast): Integer; overload;

{   G_Deflate  ,    
       ,   .
    ,      
   DestinationBytes.     
  (SourceLength + 4) .    ,   ,
    DestinationBytes.    DestinationBytes
    nil,   ""   
  .        
  . , , ,   "" 
      ,      . }

function G_Deflate(SourceBytes: Pointer; SourceLength: Integer;
  DestinationBytes: Pointer; Mode: TCompressionMode = dcmFast): Integer; overload;

{ G_ReleaseDeflator   ,  ,
       150 .  
        G_Deflate.  G_Deflate 
   ,     , 
  G_ReleaseDeflator        
    . }

procedure G_ReleaseDeflator;

{ G_GetInflatedLength      , 
    ,      SourceBytes.
  ,      G_Deflate. }

function G_GetInflatedLength(SourceBytes: Pointer): Integer;

{ G_Inflate   ,   SourceBytes,
   SourceLength .     , 
     ,     BeforeGap 
         AfterGap     .
         
  DestinationBytes.     (  BeforeGap
   AfterGap)    . }

function G_Inflate(SourceBytes: Pointer; SourceLength: Integer;
  var DestinationBytes: Pointer; BeforeGap, AfterGap: Integer): Integer; overload;

{   G_Inflate  ,    
       ,   .
     ,     ,
    DestinationBytes.     
      G_GetInflatedLength.  G_Inflate
    ,   ,  
  DestinationBytes.       nil, 
    G_GetInflatedLength, ..   
  ,     . }

function G_Inflate(SourceBytes: Pointer; SourceLength: Integer;
  DestinationBytes: Pointer): Integer; overload;

{ G_ReleaseInflator  ,   ,
     .     
   G_Inflate.      7 .
          . }

procedure G_ReleaseInflator;

implementation

{ FORMAT OF THE DATA BLOCK

  Encoded data block consists of sequences of symbols drawn from
  three conceptually distinct alphabets: either literal bytes,
  from the alphabet of byte values (0..255), or <length, backward
  distance> pairs, where the length is drawn from (3..2^21) and
  the distance is drawn from (1..262,143). In fact, the literal
  and length alphabets are merged into a single alphabet (0..286),
  where values 0..255 represent literal bytes, and values
  256..286 represent length codes (possibly in conjunction with
  extra bits following the symbol code) as follows:

      Extra               Extra                Extra
  Code Bits Length(s) Code Bits Length(s) Code  Bits   Length(s)
  ---- ---- -------   ---- ---- -------   ----  -----  -------
   256   0     3       268   0    15       280    2     41-44
   257   0     4       269   0    16       281    2     45-48
   258   0     5       270   1   17,18     282    3     49-56
   259   0     6       271   1   19,20     283    3     57-64
   260   0     7       272   1   21,22     284    4     65-80
   261   0     8       273   1   23,24     285    6     81-144
   262   0     9       274   1   25,26     286 8,16,23 145-2^21
   263   0    10       275   1   27,28
   264   0    11       276   1   29,30
   265   0    12       277   1   31,32
   266   0    13       278   2   33-36
   267   0    14       279   2   37-40

  Distance codes 0-63 are represented by codes of variable length
  with possible additional bits.

            Extra                    Extra
        Code Bits   Distance    Code  Bits    Distance
        ---- ----   --------    ----  ----    ---------
          0    0        1        32     7      513-640
          1    0        2        33     7      641-768
          2    0        3        34     7      769-896
          3    0        4        35     7      897-1024
          4    0        5        36     8     1025-1280
          5    0        6        37     8     1281-1536
          6    0        7        38     8     1537-1792
          7    0        8        39     8     1793-2048
          8    1       9,10      40     9     2049-2560
          9    1      11,12      41     9     2561-3072
         10    1      13,14      42     9     3073-3584
         11    1      15,16      43     9     3585-4096
         12    2      17-20      44    10     4097-5120
         13    2      21-24      45    10     5121-6144
         14    2      25-28      46    10     6145-7168
         15    2      29-32      47    10     7169-8192
         16    3      33-40      48    11     8193-10240
         17    3      41-48      49    11    10241-12288
         18    3      49-56      50    11    12289-14336
         19    3      57-64      51    11    14337-16384
         20    4      65-80      52    12    16385-20480
         21    4      81-96      53    12    20481-24576
         22    4      97-112     54    12    24577-28672
         23    4     113-128     55    12    28673-32768
         24    5     129-160     56    13    32769-40960
         25    5     161-192     57    13    40961-49152
         26    5     193-224     58    13    49153-57344
         27    5     225-256     59    13    57345-65536
         28    6     257-320     60    15    65537-98304
         29    6     321-384     61    15    98305-131072
         30    6     385-448     62    16   131073-196608
         31    6     449-512     63    16   196609-262143

  Fixed Huffman code lengths for the literal/length alphabet are:

       Lit Value    Bits
       ---------    ----
           0         7      1 from 128 of 7 bits (127)
         1 - 142     8      142 from 254 of 8 bits (112)
       143 - 254     9      112 from 224 of 9 bits (112)
          255        8      1 from 56 of 8 bits (55)
       256 - 279     7      24 from 27.5 of 7 bits (3.5)
       280 - 286     8      7 from 7 of 8 bits (0)

  The alphabet for code lengths is as follows:

     0 - 14: Represent code lengths of 0 - 14
       15:   Copy twice the previous code length.
       16:   Copy the previous code length 3, 4 times.
             (1 bits of length: 0 = 3, 1 = 4)
       17:   Copy the previous code length 5 - 8 times.
             (2 bits of length: 0 = 5, ... , 3 = 8)
       18:   Copy the previous code length 9 - 16 times.
             (3 bits of length: 0 = 9, ... , 7 = 16)
       19:   Copy the previous code length 17 - 144 times
             (7 bits of length)

  We can now define the format of the block:

     1 Bit: X, 1 means using of dynamic Huffman codes
               0 - see description of the next bit

     1 Bit (if X = 0): F,
            1 means using of fixed Huffman codes
            0 non-compressed block of data

     8 Bits (if X = 0 and F = 0): # of Bytes in this block:
            # - 8192, 0 if this block is the last

     if X = 1

     4 Bits: HCLEN, # of Code Length codes - 5 (5 - 20)

     (HCLEN + 5) x 3 bits: code lengths for the code length
        alphabet given just above, in the order:
        9, 10, 8, 15, 7, 11, 0, 6, 16, 5, 12, 4, 17, 13, 3,
        18, 19, 2, 1, 14

        These code lengths are interpreted as 3-bit integers
        (0-7); as above, a code length of 0 means the
        corresponding symbol (literal/length or distance code
        length) is not used.

     5 Bits: HLIT, # of Literal/Length codes - 256 (256 - 287)

     HLIT + 256 code lengths for the literal/length alphabet,
        encoded using the code length Huffman code

     6 Bits: HDIST, # of Distance codes - 1 (1 - 64)

     HDIST + 1 code lengths for the distance alphabet,
        encoded using the code length Huffman code

     1..8192..8447 literal/length and distance Huffman codes
}

uses Windows, AcedBinary, AcedConsts;

const
  ChunkShift = 15;
  ChunkCapacity = 8192;
  BlockSize = 8192;
  MaxLength = 2097296;
  Len3MaxDist = 4096;
  Len4MaxDist = 65536;
  Diff1Min = 1025;
  Diff2Min = 65537;
  LazyLimit = 32;
  CharCount = 287;
  LastChar = CharCount - 1;
  FirstLengthChar = 256;
  FirstCharWithExBit = 270;
  CharTreeSize = CharCount * 2;
  DistCount = 64;
	FirstDistWithExBit = 8;
  DistTreeSize = DistCount * 2;
  MaxBits = 14;
  ChLenCount = 20;
  ChLenTreeSize = ChLenCount * 2;
  MaxChLenBits = 7;

  HashSizeFastest = $1000;
  HashMaskFastest = $0FFF;
  PrevSizeFastest = $2000;
  PrevMaskFastest = $1FFF;
  ChainFastest = 2;
  ShiftFastest = 4;

  HashSizeFast = $8000;
  HashMaskFast = $7FFF;
  PrevSizeFast = $10000;
  PrevMaskFast = $FFFF;
  ChainFast = 8;
  ShiftFast = 5;

  HashSizeNormal = $40000;
  HashMaskNormal = $3FFFF;
  PrevSizeNormal = $20000;
  PrevMaskNormal = $1FFFF;
  ChainNormal = 32;
  ShiftNormal = 6;

  HashSizeMaximum = $40000;
  HashMaskMaximum = $3FFFF;
  PrevSizeMaximum = $40000;
  PrevMaskMaximum = $3FFFF;
  ChainMaximum = 64;
  ShiftMaximum = 6;

const
  CharExBitLength: array[0..15] of Integer =
    (1,1,1,1,1,1,1,1,2,2,2,2,3,3,4,6);

  CharExBitBase: array[0..15] of Integer =
    (17,19,21,23,25,27,29,31,33,37,41,45,49,57,65,81);

  DistExBitLength: array[0..63] of Integer =
    (0,0,0,0,0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,4,5,5,5,5,6,6,6,6,7,7,7,7,8,
     8,8,8,9,9,9,9,10,10,10,10,11,11,11,11,12,12,12,12,13,13,13,13,15,15,16,16);

  DistExBitBase: array[0..63] of Integer =
		(1,2,3,4,5,6,7,8,9,11,13,15,17,21,25,29,33,41,49,57,65,81,97,113,129,161,
     193,225,257,321,385,449,513,641,769,897,1025,1281,1537,1793,2049,2561,
     3073,3585,4097,5121,6145,7169,8193,10241,12289,14337,16385,20481,24577,
     28673,32769,40961,49153,57345,65537,98305,131073,196609);

  ChLenExBitLength: array[0..19] of Integer =
    (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,2,3,7);

  FixedCharLength: array[0..286] of Integer =
		(7,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
     8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
     8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
     8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
     8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,
     9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,
     9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,
     9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,8,
     7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8);

  FixedBitCount: array[0..14] of Integer =
    (0,0,0,0,0,0,0,25,150,112,0,0,0,0,0);

var
  _hashHead: Integer;
  _length: Integer;
  _distance: Integer;
  _prevSize: Integer;
  _prevMask: Integer;
  _shift: Integer;
  _hashSize: Integer;
  _hashMask: Integer;
  _maxChain: Integer;
  _limit: Integer;
  _breakDefOffset: Integer;
  _thisBreakOffset: Integer;
  _chain: Integer;
  _maxCode: Integer;
  _heapLen: Integer;

  _hash: PIntegerItemList;
  _prev: PIntegerItemList;
  _heap: PIntegerItemList = nil;
  _workFreq: PIntegerItemList;
  _workDad: PIntegerItemList;
  _depth: PIntegerItemList;
  _tree: PIntegerItemList;
  _nextDefCode: PIntegerItemList;

  _dstIndex: Integer;
  _chunkLength: Integer;
  _chunk: PLongWordItemList;
  _bitsDef: Integer;
  _holdDef: LongWord;

  _chunkListLength: Integer;
  _chunkListCapacity: Integer;
  _chunkList: PPointerItemList;

  _srcDefBytes: PBytes;
  _dstDefBytes: PBytes;

  _exBitCount: Integer;
  _blockStart: Integer;
  _blockEndIndex: Integer;
  _bufLength: Integer;

  _buffer: PIntegerItemList;
  _len: PIntegerItemList;
  _dist: PIntegerItemList;
  _distExBits: PIntegerItemList;

  _charBitLenLen: Integer;
  _distBitLenLen: Integer;
  _chLenBitLenLen: Integer;
  _charLenCount: Integer;
  _distLenCount: Integer;

  _charBitCount: PIntegerItemList;
  _charFreqCode: PIntegerItemList;
  _charDadLen: PIntegerItemList;
  _charBitLen: PIntegerItemList;
  _charBitLenEx: PIntegerItemList;
  _distBitCount: PIntegerItemList;
  _distFreqCode: PIntegerItemList;
  _distDadLen: PIntegerItemList;
  _distBitLen: PIntegerItemList;
  _distBitLenEx: PIntegerItemList;
  _chLenBitCount: PIntegerItemList;
  _chLenFreqCode: PIntegerItemList;
  _chLenDadLen: PIntegerItemList;
  _chLenBitLen: PIntegerItemList;

  _srcIndex: Integer;
  _break32Offset: Integer;
  _break16Offset: Integer;
  _breakInfOffset: Integer;
  _holdInf: Integer;
  _bitsInf: Integer;
  _outCounter: Integer;
  _inCounter: Integer;

  _srcInfBytes: PBytes;
  _dstInfBytes: PBytes;

  _charTree: PIntegerItemList;
  _distTree: PIntegerItemList;
  _chLenTree: PIntegerItemList;
  _bitCount: PIntegerItemList;
  _nextInfCode: PIntegerItemList;
  _bitLen: PIntegerItemList = nil;

  _fixedTree: Boolean;
  _tryMode: Boolean;
  _growMode: Boolean;

procedure NextChunk;
var
  NewList: PPointerItemList;
begin
  if _chunkListCapacity = 0 then
  begin
    _chunkListCapacity := 16;
    GetMem(_chunkList, _chunkListCapacity * SizeOf(Pointer));
    _chunkList^[0] := _chunk;
    _chunkListLength := 1;
  end else
  begin
    if _chunkListLength = _chunkListCapacity then
    begin
      _chunkListCapacity := _chunkListCapacity shl 1;
      GetMem(NewList, _chunkListCapacity * SizeOf(Pointer));
      G_CopyLongs(_chunkList, NewList, _chunkListLength);
      FreeMem(_chunkList);
      _chunkList := NewList;
    end;
    _chunkList^[_chunkListLength] := _chunk;
    Inc(_chunkListLength);
  end;
  GetMem(_chunk, ChunkCapacity * SizeOf(LongWord));
  _chunkLength := 0;
end;

function PutBit(V: Integer): Boolean;
begin
  if _bitsDef <> 0 then
  begin
    _holdDef := _holdDef or (LongWord(V) shl (32 - _bitsDef));
    Dec(_bitsDef);
    Result := True;
    Exit;
  end;
  Inc(_dstIndex, 4);
  Result := False;
  if _dstIndex <= _breakDefOffset then
  begin
    if _growMode then
    begin
      if _chunkLength = ChunkCapacity then
        NextChunk;
      _chunk^[_chunkLength] := _holdDef;
      Inc(_chunkLength);
    end
    else if not _tryMode then
    begin
      PLongWord(_dstDefBytes)^ := _holdDef;
      Inc(LongWord(_dstDefBytes), 4);
    end;
    _holdDef := LongWord(V);
    _bitsDef := 31;
    Result := True;
  end;
end;

function PutNBits(N: Integer; V: LongWord): Boolean;
var
  Bits: Integer;
begin
  Bits := _bitsDef;
  if Bits >= N then
  begin
    _holdDef := _holdDef or (V shl (32 - Bits));
    _bitsDef := Bits - N;
    Result := True;
    Exit;
  end;
  if Bits <> 0 then
  begin
    _holdDef := _holdDef or (V shl (32 - Bits));
    V := V shr Bits;
  end;
  Inc(_dstIndex, 4);
  Result := False;
  if _dstIndex <= _breakDefOffset then
  begin
    if _growMode then
    begin
      if _chunkLength = ChunkCapacity then
        NextChunk;
      _chunk^[_chunkLength] := _holdDef;
      Inc(_chunkLength);
    end
    else if not _tryMode then
    begin
      PLongWord(_dstDefBytes)^ := _holdDef;
      Inc(LongWord(_dstDefBytes), 4);
    end;
    _holdDef := V;
    _bitsDef := 32 + Bits - N;
    Result := True;
  end;
end;

procedure InitHeap(Count: Integer);
var
  H, W: PInteger;
  I: Integer;
begin
  H := @_heap^[1];
  W := @_workFreq[0];
  _heapLen := 0;
  _maxCode := -1;
  I := 0;
  while I < Count do
  begin
    if W^ <> 0 then
    begin
      H^ := I;
      _maxCode := I;
      _depth^[I] := 0;
      Inc(_heapLen);
      Inc(LongWord(H), 4);
    end else
      _workDad^[I] := 0;
    Inc(I);
    Inc(LongWord(W), 4);
  end;
  if _heapLen > 1 then
    Exit;
  if _heapLen > 0 then
  begin
    if _maxCode = 0 then
    begin
      I := 1;
      _maxCode := I;
    end else
      I := 0;
    H^ := i;
    _workFreq^[I] := 1;
    _depth^[I] := 0;
  end else
  begin
    _maxCode := 1;
    H^ := 0;
    PIntegerItemList(H)^[1] := 1;
    _workFreq^[0] := 1;
    _workFreq^[1] := 1;
    _depth^[0] := 0;
    _depth^[1] := 0;
  end;
  _heapLen := 2;
end;

procedure SortHeap(L, R: Integer);
var
  H, W: PIntegerItemList;
  I, J, M, T: Integer;
begin
  H := _heap;
  W := _workFreq;
  repeat
    I := L;
    J := R;
    M := W^[H^[(L + R) shr 1]];
    repeat
      while W^[H^[I]] < M do
        Inc(I);
      while M < W^[H^[J]] do
        Dec(J);
      if I <= J then
      begin
        T := H^[I];
        H^[I] := H^[J];
        H^[J] := T;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then
      SortHeap(L, J);
    L := I;
  until I >= R;
end;

procedure PQDownHeap;
var
  H: PIntegerItemList;
  p0, n0, p1, n1: Integer;
  V, K, J, nV: Integer;
begin
  H := _heap;
  V := H^[1];
  nV := _workFreq^[V];
  K := 1;
  J := 2;
  while J <= _heapLen do
  begin
    p0 := H^[J];
    n0 := _workFreq^[p0];
    if J < _heapLen then
    begin
      p1 := H^[J + 1];
      n1 := _workFreq^[p1];
      if (n1 < n0) or ((n1 = n0) and (_depth^[p1] <= _depth^[p0])) then
      begin
        Inc(J);
        p0 := p1;
        n0 := n1;
      end;
    end;
    if (nV < n0) or ((nV = n0) and (_depth^[V] <= _depth^[p0])) then
      Break;
    H^[K] := H^[J];
    K := J;
    J := J shl 1;
  end;
  H^[K] := V;
end;

function BuildTree(NextNode: Integer): Integer;
var
  I, N, M: Integer;
  T: PIntegerItemList;
  Heap1: PInteger;
begin
  I := 0;
  Heap1 := @_heap^[1];
  T := _tree;
  repeat
    N := Heap1^;
    Heap1^ := _Heap^[_heapLen];
    Dec(_heapLen);
    PQDownHeap;
    M := Heap1^;
    T^[0] := N;
    T^[1] := M;
    Inc(I, 2);
    Inc(LongWord(T), 8);
    _workDad^[M] := nextNode;
    _workDad^[N] := nextNode;
    _workFreq^[nextNode] := _workFreq^[N] + _workFreq^[M];
    N := _depth^[N];
    M := _depth^[M];
    if N > M then
      _depth^[NextNode] := N + 1
    else
      _depth^[NextNode] := M + 1;
    Heap1^ := nextNode;
    Inc(NextNode);
    PQDownHeap;
  until _heapLen = 1;
  _workDad^[Heap1^] := 0;
  Result := I;
end;

function PrepareCharLengths: Integer;
var
  N, M, I, OverF: Integer;
  T: PInteger;
begin
  _workFreq := _charFreqCode;
  _workDad := _charDadLen;
  InitHeap(CharCount);
  SortHeap(1, _heapLen);
  I := BuildTree(CharCount);
  G_FillLongs(0, _charBitCount, MaxBits + 1);
  OverF := 0;
  Result := 0;
  T := @_tree^[I];
  repeat
    Dec(LongWord(T), 4);
    N := T^;
    M := _charDadLen^[_charDadLen^[N]] + 1;
    if M > MaxBits then
    begin
      M := MaxBits;
      Inc(OverF);
    end;
    _charDadLen^[N] := M;
    if N < CharCount then
    begin
      Inc(Result, M * _charFreqCode^[N]);
      Inc(_charBitCount^[M]);
    end;
    Dec(I);
  until I = 0;
  if OverF = 0 then
    Exit;
  repeat
    I := MaxBits - 1;
    while _charBitCount^[I] = 0 do
      Dec(I);
    Dec(_charBitCount^[I]);
    Inc(_charBitCount^[I + 1], 2);
    Dec(_charBitCount^[MaxBits]);
    Dec(OverF, 2);
  until OverF <= 0;
  OverF := MaxBits;
  repeat
    N := _charBitCount^[OverF];
    while N <> 0 do
    begin
      M := T^;
      Inc(LongWord(T), 4);
      if M < CharCount then
      begin
        Inc(Result, (OverF - _charDadLen^[M]) * _charFreqCode^[M]);
        _charDadLen^[M] := OverF;
        Dec(N);
      end;
    end;
    Dec(OverF);
  until OverF = 0;
end;

function PrepareDistLengths: Integer;
var
  N, M, I, OverF: Integer;
  T: PInteger;
begin
  _workFreq := _distFreqCode;
  _workDad := _distDadLen;
  InitHeap(DistCount);
  SortHeap(1, _heapLen);
  I := BuildTree(DistCount);
  G_FillLongs(0, _distBitCount, MaxBits + 1);
  OverF := 0;
  Result := 0;
  T := @_tree[I];
  repeat
    Dec(LongWord(T), 4);
    N := T^;
    M := _distDadLen^[_distDadLen^[N]] + 1;
    if M > MaxBits then
    begin
      M := MaxBits;
      Inc(OverF);
    end;
    _distDadLen^[N] := M;
    if N < DistCount then
    begin
      Inc(Result, M * _distFreqCode^[N]);
      Inc(_distBitCount^[M]);
    end;
    Dec(I);
  until I = 0;
  if OverF = 0 then
    Exit;
  repeat
    I := MaxBits - 1;
    while _distBitCount^[I] = 0 do
      Dec(I);
    Dec(_distBitCount^[I]);
    Inc(_distBitCount^[I + 1], 2);
    Dec(_distBitCount^[MaxBits]);
    Dec(OverF, 2);
  until OverF <= 0;
  OverF := MaxBits;
  repeat
    N := _distBitCount^[OverF];
    while N <> 0 do
    begin
      M := T^;
      Inc(LongWord(T), 4);
      if M < DistCount then
      begin
        Inc(Result, (OverF - _distDadLen^[M]) * _distFreqCode^[M]);
        _distDadLen^[M] := OverF;
        Dec(N);
      end;
    end;
    Dec(OverF);
  until OverF = 0;
end;

function FillCharBitLengths: Integer;
var
  Dad, Len: PInteger;
  LastLen, M, I, J, BreakOffset: Integer;
begin
  Dad := PInteger(_charDadLen);
  Len := PInteger(_charBitLen);
  _charBitLenLen := 0;
  LastLen := 7;
  I := 0;
  if _maxCode >= 255 then
    _charLenCount := _maxCode + 1
  else
  begin
    _charLenCount := 256;
    _maxCode := 255;
  end;
  G_FillLongs(0, _charBitLenEx, _charLenCount);
  Result := 0;
  repeat
    M := Dad^;
    Inc(LongWord(Dad), 4);
    if (M <> LastLen) or (M <> Dad^) then
    begin
      LastLen := M;
      Len^ := M;
      Inc(I);
    end else
    begin
      if I + 144 > _maxCode then
        BreakOffset := _charLenCount
      else
        BreakOffset := I + 144;
      J := I;
      Inc(I, 2);
      Inc(LongWord(Dad), 4);
      while (I < BreakOffset) and (M = Dad^) do
      begin
        Inc(LongWord(Dad), 4);
        Inc(I);
      end;
      J := I - J;
      if J = 2 then
        Len^ := 15
      else if J < 5 then
      begin
        Len^ := 16;
        _charBitLenEx^[_charBitLenLen] := J - 3;
        Inc(Result);
      end
      else if J < 9 then
      begin
        Len^ := 17;
        _charBitLenEx^[_charBitLenLen] := J - 5;
        Inc(Result, 2);
      end
      else if J < 17 then
      begin
        Len^ := 18;
        _charBitLenEx^[_charBitLenLen] := J - 9;
        Inc(Result, 3);
      end else
      begin
        Len^ := 19;
        _charBitLenEx^[_charBitLenLen] := J - 17;
        Inc(Result, 7);
      end;
    end;
    Inc(_charBitLenLen);
    Inc(LongWord(Len), 4);
  until I >= _maxCode;
  if I = _maxCode then
  begin
    Len^ := Dad^;
    Inc(_charBitLenLen);
  end;
end;

function FillDistBitLengths: Integer;
var
  Dad, Len: PInteger;
  LastLen, M, I, J: Integer;
begin
  Dad := PInteger(_distDadLen);
  Len := PInteger(_distBitLen);
  _distBitLenLen := 0;
  LastLen := 7;
  I := 0;
  _distLenCount := _maxCode + 1;
  G_FillLongs(0, _distBitLenEx, _distLenCount);
  Result := 0;
  repeat
    M := Dad^;
    Inc(LongWord(Dad), 4);
    if (M <> LastLen) or (M <> Dad^) then
    begin
      LastLen := M;
      Len^ := M;
      Inc(I);
    end
    else
    begin
      J := I;
      Inc(I, 2);
      Inc(LongWord(Dad), 4);
      while (I < _distLenCount) and (M = Dad^) do
      begin
        Inc(LongWord(Dad), 4);
        Inc(I);
      end;
      J := I - J;
      if J = 2 then
        Len^ := 15
      else if J < 5 then
      begin
        Len^ := 16;
        _distBitLenEx^[_distBitLenLen] := J - 3;
        Inc(Result);
      end
      else if J < 9 then
      begin
        Len^ := 17;
        _distBitLenEx^[_distBitLenLen] := J - 5;
        Inc(Result, 2);
      end
      else if J < 17 then
      begin
        Len^ := 18;
        _distBitLenEx^[_distBitLenLen] := J - 9;
        Inc(Result, 3);
      end else
      begin
        Len^ := 19;
        _distBitLenEx^[_distBitLenLen] := J - 17;
        Inc(Result, 7);
      end;
    end;
    Inc(_distBitLenLen);
    Inc(LongWord(Len), 4);
  until I >= _maxCode;
  if I = _maxCode then
  begin
    Len^ := Dad^;
    Inc(_distBitLenLen);
  end;
end;

procedure FillChLenFreqCodes;
var
  Freq, Len: PIntegerItemList;
  I: Integer;
begin
  Freq := _chLenFreqCode;
  G_FillLongs(0, Freq, ChLenCount);
  Len := _charBitLen;
  I := 7;
  while I < _charBitLenLen do
  begin
    Inc(Freq^[Len^[0]]);
    Inc(Freq^[Len^[1]]);
    Inc(Freq^[Len^[2]]);
    Inc(Freq^[Len^[3]]);
    Inc(Freq^[Len^[4]]);
    Inc(Freq^[Len^[5]]);
    Inc(Freq^[Len^[6]]);
    Inc(Freq^[Len^[7]]);
    Inc(LongWord(Len), 32);
    Inc(I, 8);
  end;
  Dec(I, 7);
  while I < _charBitLenLen do
  begin
    Inc(Freq^[PInteger(Len)^]);
    Inc(LongWord(Len), 4);
    Inc(I);
  end;
  Len := _distBitLen;
  I := 7;
  while i < _distBitLenLen do
  begin
    Inc(Freq^[Len^[0]]);
    Inc(Freq^[Len^[1]]);
    Inc(Freq^[Len^[2]]);
    Inc(Freq^[Len^[3]]);
    Inc(Freq^[Len^[4]]);
    Inc(Freq^[Len^[5]]);
    Inc(Freq^[Len^[6]]);
    Inc(Freq^[Len^[7]]);
    Inc(LongWord(Len), 32);
    Inc(I, 8);
  end;
  Dec(I, 7);
  while I < _distBitLenLen do
  begin
    Inc(Freq^[PInteger(Len)^]);
    Inc(LongWord(Len), 4);
    Inc(I);
  end;
end;

function PrepareChLenLengths: Integer;
var
  N, M, I, OverF: Integer;
  T: PInteger;
begin
  _workFreq := _chLenFreqCode;
  _workDad := _chLenDadLen;
  InitHeap(ChLenCount);
  SortHeap(1, _heapLen);
  I := BuildTree(ChLenCount);
  G_FillLongs(0, _chLenBitCount, MaxChLenBits + 1);
  OverF := 0;
  Result := 0;
  T := @_tree[I];
  repeat
    Dec(LongWord(T), 4);
    N := T^;
    M := _chLenDadLen^[_chLenDadLen^[N]] + 1;
    if M > MaxChLenBits then
    begin
      M := MaxChLenBits;
      Inc(OverF);
    end;
    _chLenDadLen^[N] := M;
    if N < ChLenCount then
    begin
      Inc(Result, M * _chLenFreqCode^[N]);
      Inc(_chLenBitCount^[M]);
    end;
    Dec(I);
  until I = 0;
  if OverF = 0 then
    Exit;
  repeat
    I := MaxChLenBits - 1;
    while _chLenBitCount^[I] = 0 do
      Dec(I);
    Dec(_chLenBitCount^[I]);
    Inc(_chLenBitCount^[I + 1], 2);
    Dec(_chLenBitCount^[MaxChLenBits]);
    Dec(OverF, 2);
  until OverF <= 0;
  OverF := MaxChLenBits;
  repeat
    N := _chLenBitCount^[OverF];
    while N <> 0 do
    begin
      M := T^;
      Inc(LongWord(T), 4);
      if M < ChLenCount then
      begin
        Inc(Result, (OverF - _chLenDadLen^[M]) * _chLenFreqCode^[M]);
        _chLenDadLen^[M] := OverF;
        Dec(N);
      end;
    end;
    Dec(OverF);
  until OverF = 0;
end;

function FillChLenBitLengths: Integer;
var
  Dad, Len: PIntegerItemList;
begin
  Dad := _chLenDadLen;
  Len := _chLenBitLen;
  Len^[0] := Dad^[9];
  Len^[1] := Dad^[10];
  Len^[2] := Dad^[8];
  Len^[3] := Dad^[15];
  Len^[4] := Dad^[7];
  Len^[5] := Dad^[11];
  Len^[6] := Dad^[0];
  Len^[7] := Dad^[6];
  Len^[8] := Dad^[16];
  Len^[9] := Dad^[5];
  Len^[10] := Dad^[12];
  Len^[11] := Dad^[4];
  Len^[12] := Dad^[17];
  Len^[13] := Dad^[13];
  Len^[14] := Dad^[3];
  Len^[15] := Dad^[18];
  Len^[16] := Dad^[19];
  Len^[17] := Dad^[2];
  Len^[18] := Dad^[1];
  Len^[19] := Dad^[14];
  Result := ChLenCount - 1;
  while (Len^[Result] = 0) and (Result > 5) do
    Dec(Result);
  Inc(Result);
  _chLenBitLenLen := Result;
  Result := Result * 3;
end;

function GetFixedBitAmount: Integer;
var
  F, L: PIntegerItemList;
  I: Integer;
begin
  F := _distFreqCode;
  Result := F^[0]+F^[1]+F^[2]+F^[3]+F^[4]+F^[5]+F^[6]+F^[7]+F^[8]+
    F^[9]+F^[10]+F^[11]+F^[12]+F^[13]+F^[14]+F^[15];
  Inc(Result, F^[16]+F^[17]+F^[18]+F^[19]+F^[20]+F^[21]+F^[22]+F^[23]+
    F^[24]+F^[25]+F^[26]+F^[27]+F^[28]+F^[29]+F^[30]+F^[31]);
  Inc(Result, F^[32]+F^[33]+F^[34]+F^[35]+F^[36]+F^[37]+F^[38]+F^[39]+
    F^[40]+F^[41]+F^[42]+F^[43]+F^[44]+F^[45]+F^[46]+F^[47]);
  Inc(Result, F^[48]+F^[49]+F^[50]+F^[51]+F^[52]+F^[53]+F^[54]+F^[55]+
    F^[56]+F^[57]+F^[58]+F^[59]+F^[60]+F^[61]+F^[62]+F^[63]);
  Result := Result * 6;
  L := @FixedCharLength;
  F := _charFreqCode;
  I := 7;
  while I < CharCount do
  begin
    Inc(Result, L^[0] * F^[0]);
    Inc(Result, L^[1] * F^[1]);
    Inc(Result, L^[2] * F^[2]);
    Inc(Result, L^[3] * F^[3]);
    Inc(Result, L^[4] * F^[4]);
    Inc(Result, L^[5] * F^[5]);
    Inc(Result, L^[6] * F^[6]);
    Inc(Result, L^[7] * F^[7]);
    Inc(LongWord(L), 32);
    Inc(LongWord(F), 32);
    Inc(I, 8);
  end;
  Dec(I, 7);
  while I < CharCount do
  begin
    Inc(Result, PInteger(L)^ * PInteger(F)^);
    Inc(LongWord(L), 4);
    Inc(LongWord(F), 4);
    Inc(I);
  end;
  Inc(Result, 2);
end;

function WriteNonCompressedBlock(V: LongWord; BlockStart, BlockEnd: Integer): Boolean;
var
  Bits: Integer;
begin
  Result := False;
  if not PutBit(0) then
    Exit;
  if not PutBit(0) then
    Exit;
  Bits := _bitsDef;
  repeat
    if Bits >= 8 then
    begin
      _holdDef := _holdDef or (V shl (32 - Bits));
      Dec(Bits, 8);
    end else
    begin
      if Bits <> 0 then
      begin
        _holdDef := _holdDef or (V shl (32 - Bits));
        V := V shr Bits;
      end;
      Inc(_dstIndex, 4);
      if _dstIndex > _breakDefOffset then
        Exit;
      if _growMode then
      begin
        if _chunkLength = ChunkCapacity then
          NextChunk;
        _chunk^[_chunkLength] := _holdDef;
        Inc(_chunkLength);
      end
      else if not _tryMode then
      begin
        PLongWord(_dstDefBytes)^ := _holdDef;
        Inc(LongWord(_dstDefBytes), 4);
      end;
      _holdDef := V;
      Inc(Bits, 24);
    end;
    if BlockStart < BlockEnd then
      V := _srcDefBytes^[BlockStart]
    else
      Break;
    Inc(BlockStart);
  until False;
  _bitsDef := Bits;
  Result := True;
end;

function WriteBlockData: Boolean;
label
  NextLoop;
var
  I, C, N: Integer;
begin
  Result := False;
  I := 0;
  while I < _bufLength do
  begin
    C := _buffer^[I];
    if not PutNBits(_charDadLen^[C], LongWord(_charFreqCode^[C])) then
      Exit;
    if C < FirstLengthChar then
      goto NextLoop;
    if C >= FirstCharWithExBit then
      if C <> LastChar then
      begin
        Dec(C, FirstCharWithExBit);
        if not PutNBits(CharExBitLength[C], LongWord(_len^[I] - CharExBitBase[C])) then
          Exit;
      end else
      begin
        C := _len[I] - 145;
        if C < 128 then
        begin
          if not PutNBits(8, LongWord(C or $80)) then
            Exit;
        end
        else if C < 16384 then
        begin
          if not PutNBits(16, LongWord((C and $7F) or
            ((C and $3F80) shl 1) or $8000)) then
            Exit;
        end
        else if not PutNBits(23, LongWord((C and $7F) or
          ((C and $3F80) shl 1) or ((C and $1FC000) shl 2))) then
          Exit;
      end;
    C := _dist^[I];
    N := _distDadLen^[C];
    if not PutNBits(N + DistExBitLength[C], LongWord(_distFreqCode^[C] or
      (_distExBits^[I] shl N))) then
      Exit;
  NextLoop:
    Inc(I);
  end;
  Result := True;
end;

procedure GenerateFixedCodes;
var
  P: PIntegerItemList;
  BD, FC: PInteger;
  N, I, M: Integer;
begin
  P := _nextDefCode;
  P^[1] := 0;
  N := FixedBitCount[1] shl 1;
  P^[2] := N;
  BD := @FixedBitCount[2];
  Inc(LongWord(P), 12);
  I := MaxBits - 2;
  while I > 0 do
  begin
    N := (N + BD^) shl 1;
    PInteger(P)^ := N;
    Inc(LongWord(BD), 4);
    Inc(LongWord(P), 4);
    Dec(I);
  end;
  P := _nextDefCode;
  FC := PInteger(_charFreqCode);
  BD := @FixedCharLength[0];
  I := CharCount;
  while I > 0 do
  begin
    N := BD^;
    M := P^[N];
    FC^ := Integer(G_ReverseBits(LongWord(M), N));
    Inc(LongWord(BD), 4);
    P^[N] := M + 1;
    Inc(LongWord(FC), 4);
    Dec(I);
  end;
  G_CopyLongs(@FixedCharLength, _charDadLen, CharCount);
  FC := PInteger(_distFreqCode);
  I := 0;
  while I < DistCount do
  begin
    FC^ := I;
    Inc(LongWord(FC), 4);
    Inc(I);
  end;
  G_FillLongs(6, _distDadLen, DistCount);
end;

function WriteFixedBlock: Boolean;
begin
  GenerateFixedCodes;
  Result := False;
  if not PutBit(0) then
    Exit;
  if not  PutBit(1) then
    Exit;
  Result := WriteBlockData();
end;

procedure GenerateCodes;
var
  P: PIntegerItemList;
  BD, FC: PInteger;
  N, I, M: Integer;
begin
  P := _nextDefCode;
  P^[1] := 0;
  N := _chLenBitCount^[1] shl 1;
  P^[2] := N;
  Inc(LongWord(P), 12);
  BD := @_chLenBitCount^[2];
  I := MaxChLenBits - 2;
  while i > 0 do
  begin
    N := (N + BD^) shl 1;
    Inc(LongWord(BD), 4);
    PInteger(P)^ := N;
    Inc(LongWord(P), 4);
    Dec(I);
  end;
  P := _nextDefCode;
  FC := PInteger(_chLenFreqCode);
  BD := PInteger(_chLenDadLen);
  I := ChLenCount;
  while I > 0 do
  begin
    N := BD^;
    if N <> 0 then
    begin
      M := P^[N];
      FC^ := Integer(G_ReverseBits(LongWord(M), N));
      P^[N] := M + 1;
    end;
    Inc(LongWord(BD), 4);
    Inc(LongWord(FC), 4);
    Dec(I);
  end;
  P := _nextDefCode;
  P^[1] := 0;
  N := _charBitCount^[1] shl 1;
  P^[2] := N;
  Inc(LongWord(P), 12);
  BD := @_charBitCount^[2];
  I := MaxBits - 2;
  while I > 0 do
  begin
    N := (N + BD^) shl 1;
    Inc(LongWord(BD), 4);
    PInteger(P)^ := N;
    Inc(LongWord(P), 4);
    Dec(I);
  end;
  P := _nextDefCode;
  FC := PInteger(_charFreqCode);
  BD := PInteger(_charDadLen);
  I := CharCount;
  while I > 0 do
  begin
    N := BD^;
    if N <> 0 then
    begin
      M := P^[N];
      FC^ := Integer(G_ReverseBits(LongWord(M), N));
      P^[N] := M + 1;
    end;
    Inc(LongWord(BD), 4);
    Inc(LongWord(FC), 4);
    Dec(I);
  end;
  P := _nextDefCode;
  P^[1] := 0;
  N := _distBitCount^[1] shl 1;
  P^[2] := N;
  Inc(LongWord(P), 12);
  BD := @_distBitCount^[2];
  I := MaxBits - 2;
  while I > 0 do
  begin
    N := (N + BD^) shl 1;
    Inc(LongWord(BD), 4);
    PInteger(P)^ := N;
    Inc(LongWord(P), 4);
    Dec(I);
  end;
  P := _nextDefCode;
  FC := PInteger(_distFreqCode);
  BD := PInteger(_distDadLen);
  I := DistCount;
  while I > 0 do
  begin
    N := BD^;
    if N <> 0 then
    begin
      M := P^[N];
      FC^ := Integer(G_ReverseBits(LongWord(M), N));
      P^[N] := M + 1;
    end;
    Inc(LongWord(BD), 4);
    Inc(LongWord(FC), 4);
    Dec(I);
  end;
end;

function WriteDynamicBlock: Boolean;
var
  I, M, N: Integer;
begin
  GenerateCodes;
  Result := False;
  if not PutBit(1) then
    Exit;
  if not PutNBits(4, LongWord(_chLenBitLenLen - 5)) then
    Exit;
  I := 0;
  while I < _chLenBitLenLen do
  begin
    if not PutNBits(3, LongWord(_chLenBitLen[I])) then
      Exit;
    Inc(I);
  end;
  if not PutNBits(5, LongWord(_charLenCount - 256)) then
    Exit;
  I := 0;
  while I < _charBitLenLen do
  begin
    M := _charBitLen^[I];
    N := _chLenDadLen^[M];
    if not PutNBits(N + ChLenExBitLength[M],
      LongWord(_chLenFreqCode^[M] or (_charBitLenEx^[I] shl N))) then
      Exit;
    Inc(I);
  end;
  if not PutNBits(6, LongWord(_distLenCount - 1)) then
    Exit;
  I := 0;
  while I < _distBitLenLen do
  begin
    M := _distBitLen^[I];
    N := _chLenDadLen^[M];
    if not PutNBits(N + ChLenExBitLength[M],
      LongWord(_chLenFreqCode^[M] or (_distBitLenEx^[I] shl N))) then
      Exit;
    Inc(I);
  end;
  Result := WriteBlockData;
end;

function FlushBuffer: Boolean;
var
  BitAmount, N: Integer;
  FixedTree: Boolean;
begin
  BitAmount := 16;
  Inc(BitAmount, PrepareCharLengths);
  Inc(BitAmount, FillCharBitLengths);
  Inc(BitAmount, PrepareDistLengths);
  Inc(BitAmount, FillDistBitLengths);
  FillChLenFreqCodes;
  Inc(BitAmount, PrepareChLenLengths);
  Inc(BitAmount, FillChLenBitLengths);
  FixedTree := False;
  N := GetFixedBitAmount;
  if N <= BitAmount then
  begin
    BitAmount := N;
    FixedTree := True;
  end;
  Inc(BitAmount, _exBitCount);
  N := _blockEndIndex - _blockStart;
  Result := False;
  if (N <= BlockSize + 255) and ((N shl 3) + 10 <= BitAmount) then
  begin
    if N < BlockSize then
      N := 0
    else
      Dec(N, BlockSize);
    if not WriteNonCompressedBlock(LongWord(N), _blockStart, _blockEndIndex) then
      Exit;
  end
  else if not FixedTree then
  begin
    if not WriteDynamicBlock then
      Exit;
  end
  else if not WriteFixedBlock then
    Exit;
  G_FillLongs(0, _charFreqCode, CharCount);
  G_FillLongs(0, _distFreqCode, DistCount);
  _exBitCount := 0;
  _blockStart := _blockEndIndex;
  _bufLength := 0;
  Result := True;
end;

function GetLenChar(L: Integer): Integer;
begin
  if L < 17 then
  begin
    Result := L + 253;
    Exit;
  end;
  if L < 49 then
  begin
    if L < 33 then
    begin
      Inc(_exBitCount);
      Result := ((L - 17) shr 1) + 270;
      Exit;
    end;
    Inc(_exBitCount, 2);
    Result := ((L - 33) shr 2) + 278;
    Exit;
  end;
  if L < 81 then
  begin
    if L < 65 then
    begin
      Inc(_exBitCount, 3);
      Result := ((L - 49) shr 3) + 282;
      Exit;
    end;
    Inc(_exBitCount, 4);
    Result := 284;
    Exit;
  end;
  if L < 145 then
  begin
    Inc(_exBitCount, 6);
    Result := 285;
    Exit;
  end;
  if L < 273 then
    Inc(_exBitCount, 8)
  else if L < 16529 then
    Inc(_exBitCount, 16)
  else
    Inc(_exBitCount, 23);
  Result := 286;
end;

function SplitDistance(D: Integer): Integer;
begin
  if D < 1025 then
  begin
    if D < 65 then
    begin
      if D < 17 then
      begin
        if D < 9 then
          Result := D - 1
        else
          Result := ((D - 9) shr 1) + 8;
      end
      else if D < 33 then
        Result := ((D - 17) shr 2) + 12
      else
        Result := ((D - 33) shr 3) + 16;
    end
    else if D < 257 then
    begin
      if D < 129 then
        Result := ((D - 65) shr 4) + 20
      else
        Result := ((D - 129) shr 5) + 24;
    end
    else if D < 513 then
      Result := ((D - 257) shr 6) + 28
    else
      Result := ((D - 513) shr 7) + 32;
  end
  else if D < 16385 then
  begin
    if D < 4097 then
    begin
      if D < 2049 then
        Result := ((D - 1025) shr 8) + 36
      else
        Result := ((D - 2049) shr 9) + 40;
    end
    else if D < 8193 then
      Result := ((D - 4097) shr 10) + 44
    else
      Result := ((D - 8193) shr 11) + 48;
  end
  else if D < 65537 then
  begin
    if D < 32769 then
      Result := ((D - 16385) shr 12) + 52
    else
      Result := ((D - 32769) shr 13) + 56;
  end
  else if D < 131073 then
    Result := ((D - 65537) shr 15) + 60
  else
    Result := ((D - 131073) shr 16) + 62;
end;

function PutChar(C: Integer): Boolean;
var
  BufPos: Integer;
begin
  BufPos := _bufLength;
  if BufPos = BlockSize then
  begin
    if not FlushBuffer then
    begin
      Result := False;
      Exit;
    end;
    BufPos := 0;
  end;
  Inc(_charFreqCode^[C]);
  _buffer^[BufPos] := C;
  _bufLength := BufPos + 1;
  Result := True;
end;

function PutLenDist: Boolean;
var
  BufPos, C: Integer;
begin
  BufPos := _bufLength;
  if BufPos = BlockSize then
  begin
    if not FlushBuffer then
    begin
      Result := False;
      Exit;
    end;
    BufPos := 0;
  end;
  C := GetLenChar(_length);
  Inc(_charFreqCode^[C]);
  _buffer^[BufPos] := C;
  _len^[BufPos] := _length;
  C := SplitDistance(_distance);
  _distExBits^[BufPos] := _distance - DistExBitBase[C];
  Inc(_distFreqCode^[C]);
  _dist^[BufPos] := C;
  Inc(_exBitCount, DistExBitLength[C]);
  _bufLength := BufPos + 1;
  Result := True;
end;

function Test1Difference(NewDistance: Integer): Boolean;
begin
  Result := True;
  if NewDistance < 16385 then
  begin
    if NewDistance < 4097 then
    begin
      if NewDistance < 2049 then
      begin
        if _distance < 9 then
          Result := False;
      end
      else if _distance < 17 then
        Result := False;
    end
    else if NewDistance < 8193 then
    begin
      if _distance < 33 then
        Result := False;
    end
    else if _distance < 65 then
      Result := False;
  end
  else if NewDistance < 65537 then
  begin
    if NewDistance < 32769 then
    begin
      if _distance < 129 then
        Result := False;
    end
    else if _distance < 257 then
      Result := False;
  end
  else if NewDistance < 131073 then
  begin
    if _distance < 1025 then
      Result := False;
  end
  else if _distance < 2049 then
    Result := False;
end;

function Test2Difference(NewDistance: Integer): Boolean;
begin
  Result := True;
  if NewDistance < 131073 then
  begin
    if _distance < 17 then
      Result := False;
  end
  else if _distance < 33 then
    Result := False;
end;

function TestMatch(P: Pointer; M1, M2: Integer): Integer;
asm
        PUSH  EBX
        PUSH  ESI
        MOV   EBX,[_thisBreakOffset]
        LEA   ESI,[EBX-15]
        PUSH  EBX
        JMP   @@ts
@@lp1:  MOV   EBX,[EAX+EDX]
        XOR   EBX,[EAX+ECX]
        JNE   @@m0
        MOV   EBX,[EAX+EDX+4]
        XOR   EBX,[EAX+ECX+4]
        JNE   @@m1
        MOV   EBX,[EAX+EDX+8]
        XOR   EBX,[EAX+ECX+8]
        JNE   @@m2
        MOV   EBX,[EAX+EDX+12]
        XOR   EBX,[EAX+ECX+12]
        JNE   @@m3
        ADD   ECX,16
        ADD   EDX,16
@@ts:   CMP   ECX,ESI
        JL    @@lp1
        POP   ESI
        JMP   @@nx
@@m3:   ADD   ECX,4
@@m2:   ADD   ECX,4
@@m1:   ADD   ECX,4
@@m0:   TEST  EBX,$FFFF
        JNE   @@nz
        ADD   ECX,2
        SHR   EBX,16
@@nz:   SUB   BL,1
        ADC   ECX,0
        POP   ESI
        JMP   @@qt 
@@lp2:  MOV   BL,[EAX+EDX]
        CMP   BL,[EAX+ECX]
        JNE   @@qt
        INC   ECX
        INC   EDX
@@nx:   CMP   ECX,ESI
        JL    @@lp2
@@qt:   MOV   EAX,ECX
        POP   ESI
        POP   EBX
end;

function FindNextFragment(SrcIndex: Integer): Boolean;
var
  P: PBytes;
  M2, Head, Length, Chain: Integer;
  M: LongWord;
begin
  P := _srcDefBytes;
  Head := _hashHead;
  Length := _length;
  M := PLongWord(@P^[SrcIndex + Length - 3])^;
  if SrcIndex + MaxLength <= _breakDefOffset then
    _thisBreakOffset := SrcIndex + MaxLength
  else
    _thisBreakOffset := _breakDefOffset;
  Result := False;
  Chain := _chain;
  repeat
    if PLongWord(@P^[Head + Length - 3])^ = M then
    begin
      M2 := TestMatch(P, Head, SrcIndex);
      Dec(M2, SrcIndex + Length);
      Dec(SrcIndex, Head);
      if (M2 > 2) or ((M2 = 1) and ((SrcIndex < Diff1Min) or Test1Difference(SrcIndex))) or
        ((M2 = 2) and ((SrcIndex < Diff2Min) or Test2Difference(SrcIndex))) then
      begin
        Inc(Length, M2);
        _distance := SrcIndex;
        M2 := Length + SrcIndex + Head;
        Result := True;
        if M2 = _thisBreakOffset then
          Break;
        M := PLongWord(@P^[M2 - 3])^;
      end;
      Inc(SrcIndex, Head);
    end;
    Head := _prev^[Head and _prevMask];
    if Head <= _limit then
      Break;
    Dec(Chain);
  until Chain = 0;
  _length := Length;
end;

function FindFirstFragment(SrcIndex: Integer): Boolean;
var
  P: PBytes;
  Head, Chain: Integer;
  M: LongWord;
begin
  P := _srcDefBytes;
  M := LongWord(P^[SrcIndex]) or (LongWord(P^[SrcIndex + 1]) shl 8) or
    (LongWord(P^[SrcIndex + 2]) shl 16);
  Head := _hashHead;
  if SrcIndex + MaxLength <= _breakDefOffset then
    _thisBreakOffset := SrcIndex + MaxLength
  else
    _thisBreakOffset := _breakDefOffset;
  Chain := _chain;
  Result := False;
  repeat
    if PLongWord(@P^[Head])^ and $FFFFFF = M then
    begin
      if (SrcIndex + 3 < _thisBreakOffset) and (P^[SrcIndex + 3] = P^[Head + 3]) then
      begin
        if (SrcIndex + 4 < _thisBreakOffset) and (P^[SrcIndex + 4] = P^[Head + 4]) then
        begin
          _length := TestMatch(P, Head + 5, SrcIndex + 5) - SrcIndex;
          Result := True;
          Break;
        end
        else if SrcIndex - Head <= Len4MaxDist then
        begin
          _length := 4;
          Result := True;
          Break;
        end;
      end
      else if SrcIndex - Head <= Len3MaxDist then
      begin
        _length := 3;
        Result := True;
        Break;
      end;
    end;
    Head := _prev^[Head and _prevMask];
    if Head <= _limit then
      Break;
    Dec(Chain);
  until Chain = 0;
  _distance := SrcIndex - Head;
  _hashHead := Head;
  _chain := Chain;
end;

function FlushBitTrail: Boolean;
var
  Hold: LongWord;
begin
  Result := False;
  if not FlushBuffer then
    Exit;
  Inc(_dstIndex, 4 - (_bitsDef shr 3));
  if _dstIndex <= _breakDefOffset then
  begin
    Hold := _holdDef;
    if _growMode then
    begin
      if _chunkLength = ChunkCapacity then
        NextChunk;
      _chunk^[_chunkLength] := Hold;
      Inc(_chunkLength);
    end
    else if not _tryMode then
      case _bitsDef shr 3 of
        0: PLongWord(_dstDefBytes)^ := Hold;
        1:
          begin
            _dstDefBytes^[0] := Byte(Hold);
            _dstDefBytes^[1] := Byte(Hold shr 8);
            _dstDefBytes^[2] := Byte(Hold shr 16);
          end;
        2:
          begin
            _dstDefBytes^[0] := Byte(Hold);
            _dstDefBytes^[1] := Byte(Hold shr 8);
          end;
        3: PByte(_dstDefBytes)^ := Byte(Hold);
      end;
    Result := True;
  end;
end;

function DeflateCore: Integer;
var
  H, SrcIndex, NextOffset: Integer;
begin
  GetMem(_hash, _hashSize * SizeOf(Integer));
  GetMem(_prev, _prevSize * SizeOf(Integer));
  G_FillLongs(-_prevSize, _hash, _hashSize);
  G_FillLongs(0, _charFreqCode, CharCount);
  G_FillLongs(0, _distFreqCode, DistCount);
  _blockStart := 0;
  _bufLength := 0;
  _exBitCount := 0;
  _prev^[0] := -_prevSize;
  H := ((_srcDefBytes^[0] shl (_shift + _shift)) xor
    (_srcDefBytes^[1] shl _shift) xor _srcDefBytes^[2]) and _hashMask;
  _hash^[H] := 0;
  _dstIndex := 0;
  PutChar(_srcDefBytes^[0]);
  _bitsDef := 32;
  _holdDef := 0;
  SrcIndex := 1;
  Result := -1;
  while SrcIndex < _breakDefOffset - 2 do
  begin
    H := ((H shl _shift) xor _srcDefBytes^[SrcIndex + 2]) and _hashMask;
    _hashHead := _hash^[H];
    _prev^[SrcIndex and _prevMask] := _hashHead;
    _hash^[H] := SrcIndex;
    _blockEndIndex := SrcIndex;
    _chain := _maxChain;
    _limit := SrcIndex - _prevSize;
    if (_hashHead <= _limit) or not FindFirstFragment(SrcIndex) then
    begin
      if not PutChar(_srcDefBytes^[SrcIndex]) then
        Exit;
      Inc(SrcIndex);
    end else
    begin
      NextOffset := SrcIndex + _length;
      _hashHead := _prev^[_hashHead and _prevMask];
      if (_hashHead > _limit) and (NextOffset < _thisBreakOffset) then
      begin
        FindNextFragment(SrcIndex);
        NextOffset := SrcIndex + _length;
      end;
      while (_length < LazyLimit) and (NextOffset < _breakDefOffset) do
      begin
        Inc(SrcIndex);
        H := ((H shl _shift) xor _srcDefBytes^[SrcIndex + 2]) and _hashMask;
        _hashHead := _hash^[H];
        _prev^[SrcIndex and _prevMask] := _hashHead;
        _hash^[H] := SrcIndex;
        _chain := _maxChain;
        _limit := SrcIndex - _prevSize;
        if (_hashHead <= _limit) or not FindNextFragment(SrcIndex) then
          Break;
        if not PutChar(_srcDefBytes^[SrcIndex - 1]) then
          Exit;
        NextOffset := SrcIndex + _length;
        _blockEndIndex := SrcIndex;
      end;
      if not PutLenDist then
        Exit;
      if NextOffset >= _breakDefOffset - 2 then
        SrcIndex := NextOffset
      else
      begin
        Inc(SrcIndex);
        repeat
          H := ((H shl _shift) xor _srcDefBytes^[SrcIndex + 2]) and _hashMask;
          _hashHead := _hash^[H];
          _prev^[SrcIndex and _prevMask] := _hashHead;
          _hash^[H] := SrcIndex;
          Inc(SrcIndex);
        until SrcIndex = NextOffset;
      end;
    end;
  end;
  while SrcIndex < _breakDefOffset do
  begin
    _blockEndIndex := SrcIndex;
    if not PutChar(_srcDefBytes^[SrcIndex]) then
      Exit;
    Inc(SrcIndex);
  end;
  _blockEndIndex := SrcIndex;
  if not FlushBitTrail then
    Exit;
  Result := _dstIndex;
end;

procedure SetupDeflateMode(Mode: TCompressionMode);
begin
  case Mode of
    dcmFastest:
      begin
        _hashSize := HashSizeFastest;
        _hashMask := HashMaskFastest;
        _prevSize := PrevSizeFastest;
        _prevMask := PrevMaskFastest;
        _maxChain := ChainFastest;
        _shift := ShiftFastest;
      end;
    dcmFast:
      begin
        _hashSize := HashSizeFast;
        _hashMask := HashMaskFast;
        _prevSize := PrevSizeFast;
        _prevMask := PrevMaskFast;
        _maxChain := ChainFast;
        _shift := ShiftFast;
      end;
    dcmNormal:
      begin
        _hashSize := HashSizeNormal;
        _hashMask := HashMaskNormal;
        _prevSize := PrevSizeNormal;
        _prevMask := PrevMaskNormal;
        _maxChain := ChainNormal;
        _shift := ShiftNormal;
      end;
    dcmMaximumCompression:
      begin
        _hashSize := HashSizeMaximum;
        _hashMask := HashMaskMaximum;
        _prevSize := PrevSizeMaximum;
        _prevMask := PrevMaskMaximum;
        _maxChain := ChainMaximum;
        _shift := ShiftMaximum;
      end;
  end;
end;

var
  DeflatorCriticalSection: _RTL_CRITICAL_SECTION;

procedure AllocDeflator;
begin
  if _heap = nil then
  begin
    GetMem(_charBitCount, (MaxBits + 1) * SizeOf(Integer));
    GetMem(_charFreqCode, CharTreeSize * SizeOf(Integer));
    GetMem(_charDadLen, CharTreeSize * SizeOf(Integer));
    GetMem(_charBitLen, CharCount * SizeOf(Integer));
    GetMem(_charBitLenEx, CharCount * SizeOf(Integer));
    GetMem(_distBitCount, (MaxBits + 1) * SizeOf(Integer));
    GetMem(_distFreqCode, DistTreeSize * SizeOf(Integer));
    GetMem(_distDadLen, DistTreeSize * SizeOf(Integer));
    GetMem(_distBitLen, DistCount * SizeOf(Integer));
    GetMem(_distBitLenEx, DistCount * SizeOf(Integer));
    GetMem(_chLenBitCount, (MaxChLenBits + 1) * SizeOf(Integer));
    GetMem(_chLenFreqCode, ChLenTreeSize * SizeOf(Integer));
    GetMem(_chLenDadLen, ChLenTreeSize * SizeOf(Integer));
    GetMem(_chLenBitLen, ChLenCount * SizeOf(Integer));
    GetMem(_buffer, BlockSize * SizeOf(Integer));
    GetMem(_len, BlockSize * SizeOf(Integer));
    GetMem(_dist, BlockSize * SizeOf(Integer));
    GetMem(_distExBits, BlockSize * SizeOf(Integer));
    GetMem(_nextDefCode, (MaxBits + 1) * SizeOf(Integer));
    GetMem(_tree, CharTreeSize * SizeOf(Integer));
    GetMem(_depth, CharTreeSize * SizeOf(Integer));
    GetMem(_heap, (CharCount + 1) * SizeOf(Integer));
  end;
end;

procedure FreeChunks;
var
  I: Integer;
begin
  if _chunkListCapacity > 0 then
  begin
    for I := _chunkListLength - 1 downto 0 do
      FreeMem(_chunkList^[I]);
    FreeMem(_chunkList);
  end;
  FreeMem(_chunk);
end;

function FlushNonCompressed(SourceBytes: Pointer; SourceLength: Integer;
  var DestinationBytes: Pointer; BeforeGap, AfterGap: Integer): Integer;
begin
  GetMem(DestinationBytes, BeforeGap + 4 + SourceLength + AfterGap);
  if BeforeGap > 0 then
    G_ZeroMem(DestinationBytes, BeforeGap);
  PInteger(@PBytes(DestinationBytes)^[BeforeGap])^ := -SourceLength;
  Inc(BeforeGap, 4);
  G_CopyMem(SourceBytes, @PBytes(DestinationBytes)^[BeforeGap], SourceLength);
  if AfterGap > 0 then
    G_ZeroMem(@PBytes(DestinationBytes)^[BeforeGap + SourceLength], AfterGap);
  Result := SourceLength + 4;
end;

function G_Deflate(SourceBytes: Pointer; SourceLength: Integer;
  var DestinationBytes: Pointer; BeforeGap, AfterGap: Integer;
  Mode: TCompressionMode): Integer;
var
  OutSize, I, N: Integer;
begin
  if SourceLength = 0 then
  begin
    DestinationBytes := G_AllocMem(BeforeGap + AfterGap + 4);
    Result := 4;
    Exit;
  end;
  if not Assigned(SourceBytes) then
    RaiseErrorFmt(SErrArgumentNull, 'SourceBytes', 'G_Deflate');
  if (Mode = dcmNoCompression) or (SourceLength < 4) then
  begin
    Result := FlushNonCompressed(SourceBytes, SourceLength, DestinationBytes,
      BeforeGap, AfterGap);
    Exit;
  end;
  if IsMultiThread then
    EnterCriticalSection(DeflatorCriticalSection);
  try
    AllocDeflator;
    SetupDeflateMode(Mode);
    _growMode := True;
    _tryMode := False;
    _chunkListCapacity := 0;
    _chunkListLength := 0;
    GetMem(_chunk, ChunkCapacity * SizeOf(LongWord));
    _chunkLength := 0;
    _breakDefOffset := SourceLength;
    _dstDefBytes := nil;
    _srcDefBytes := PBytes(SourceBytes);
    OutSize := DeflateCore;
    FreeMem(_hash);
    FreeMem(_prev);
    if OutSize < 0 then
    begin
      FreeChunks;
      Result := FlushNonCompressed(SourceBytes, SourceLength, DestinationBytes,
        BeforeGap, AfterGap);
    end else
    begin
      Result := OutSize + 4;
      GetMem(DestinationBytes, BeforeGap + Result + AfterGap);
      if BeforeGap > 0 then
        G_ZeroMem(DestinationBytes, BeforeGap);
      PInteger(@PBytes(DestinationBytes)^[BeforeGap])^ := SourceLength;
      Inc(BeforeGap, 4);
      if AfterGap > 0 then
        G_ZeroMem(@PBytes(DestinationBytes)^[BeforeGap + OutSize], AfterGap);
      Dec(OutSize);
      N := OutSize shr ChunkShift;
      I := 0;
      while I < N do
      begin
        G_CopyMem(_chunkList^[I], @PBytes(DestinationBytes)^[BeforeGap],
          ChunkCapacity * SizeOf(LongWord));
        Inc(BeforeGap, ChunkCapacity * SizeOf(LongWord));
        Inc(I);
      end;
      G_CopyMem(_chunk, @PBytes(DestinationBytes)^[BeforeGap],
        (OutSize and ((ChunkCapacity * SizeOf(LongWord)) - 1)) + 1);
      FreeChunks;
    end;
  finally
    if IsMultiThread then
      LeaveCriticalSection(DeflatorCriticalSection);
  end;
end;

function G_Deflate(SourceBytes: Pointer; SourceLength: Integer;
  DestinationBytes: Pointer; Mode: TCompressionMode): Integer;
var
  DstBytes: PBytes;
  TryMode: Boolean;
begin
  TryMode := False;
  if destinationBytes <> nil then
    DstBytes := PBytes(DestinationBytes)
  else
  begin
    DstBytes := nil;
    TryMode := True;
  end;
  if SourceLength = 0 then
  begin
    if not TryMode then
      PInteger(DstBytes)^ := 0;
    Result := 4;
    Exit;
  end;
  if not Assigned(SourceBytes) then
    RaiseErrorFmt(SErrArgumentNull, 'SourceBytes', 'G_Deflate');
  Result := -1;
  if (Mode <> dcmNoCompression) and (SourceLength > 3) then
  begin
    if IsMultiThread then
      EnterCriticalSection(DeflatorCriticalSection);
    try
      AllocDeflator;
      SetupDeflateMode(Mode);
      _tryMode := TryMode;
      _growMode := False;
      _breakDefOffset := SourceLength;
      _srcDefBytes := PBytes(SourceBytes);
      _dstDefBytes := @DstBytes^[4];
      if not TryMode then
        PInteger(DstBytes)^ := SourceLength;
      Result := DeflateCore;
      FreeMem(_hash);
      FreeMem(_prev);
    finally
      if IsMultiThread then
        LeaveCriticalSection(DeflatorCriticalSection);
    end;
  end;
  if Result < 0 then
  begin
    if not TryMode then
    begin
      PInteger(DstBytes)^ := -SourceLength;
      G_CopyMem(SourceBytes, @DstBytes^[4], SourceLength);
    end;
    Result := SourceLength;
  end;
  Inc(Result, 4);
end;

procedure G_ReleaseDeflator;
begin
  if IsMultiThread then
    EnterCriticalSection(DeflatorCriticalSection);
  try
    if _heap <> nil then
    begin
      FreeMem(_charBitCount);
      FreeMem(_charFreqCode);
      FreeMem(_charDadLen);
      FreeMem(_charBitLen);
      FreeMem(_charBitLenEx);
      FreeMem(_distBitCount);
      FreeMem(_distFreqCode);
      FreeMem(_distDadLen);
      FreeMem(_distBitLen);
      FreeMem(_distBitLenEx);
      FreeMem(_chLenBitCount);
      FreeMem(_chLenFreqCode);
      FreeMem(_chLenDadLen);
      FreeMem(_chLenBitLen);
      FreeMem(_buffer);
      FreeMem(_len);
      FreeMem(_dist);
      FreeMem(_distExBits);
      FreeMem(_nextDefCode);
      FreeMem(_tree);
      FreeMem(_depth);
      FreeMem(_heap);
      _heap := nil;
    end;
  finally
    if IsMultiThread then
      LeaveCriticalSection(DeflatorCriticalSection);
  end;
end;

function ProcessNBitsTail(Bits: Integer): Integer;
begin
  if _srcIndex < _breakInfOffset then
  begin
    Inc(_srcIndex);
    Result := (PByte(_srcInfBytes)^ shl Bits);
    Inc(LongWord(_srcInfBytes));
  end else
  begin
    RaiseError(SErrReadBeyondTheEndOfStream);
    Result := 0;
  end;
end;

function GetNBits(N: Integer): Integer;
var
  Hold, Bits: Integer;
begin
  Hold := _holdInf;
  Bits := _bitsInf;
  if Bits < N then
    if _srcIndex < _break16Offset then
    begin
      Inc(_srcIndex, 2);
      Hold := Hold or (PWord(_srcInfBytes)^ shl Bits);
      Inc(LongWord(_srcInfBytes), 2);
      Inc(Bits, 16);
    end else
    begin
      Hold := Hold or ProcessNBitsTail(Bits);
      Inc(Bits, 8);
    end;
  _holdInf := Hold shr N;
  _bitsInf := Bits - N;
  Result := Hold and ((1 shl N) - 1);
end;

function ProcessTail: Integer;
begin
  if _srcIndex < _breakInfOffset then
  begin
    Inc(_srcIndex);
    Result := PByte(_srcInfBytes)^;
    Inc(LongWord(_srcInfBytes));
  end else
  begin
    RaiseError(SErrReadBeyondTheEndOfStream);
    Result := 0;
  end;
end;

function GetBit: Integer;
var
  Hold: LongWord;
  Bits: Integer;
begin
  Hold := LongWord(_holdInf);
  Bits := _bitsInf;
  if Bits <> 0 then
    Dec(Bits)
  else if _srcIndex < _break32Offset then
  begin
    Inc(_srcIndex, 4);
    Hold := PLongWord(_srcInfBytes)^;
    Inc(LongWord(_srcInfBytes), 4);
    Bits := 31;
  end else
  begin
    Hold := ProcessTail;
    Bits := 7;
  end;
  _holdInf := Integer(Hold shr 1);
  _bitsInf := Bits;
  Result := Integer(Hold and 1);
end;

function GetCode(Tree: PIntegerItemList): Integer;
label
  CodeFound;
var
  Code, Hold, Bits: Integer;
begin
  Code := 1;
  Hold := _holdInf;
  Bits := _bitsInf;
  repeat
    while Bits <> 0 do
    begin
      Code := Tree^[Code + (Hold and 1)];
      Hold := Hold shr 1;
      Dec(Bits);
      if Code <= 0 then
        goto CodeFound;
    end;
    if _srcIndex < _break32Offset then
    begin
      Inc(_srcIndex, 4);
      Hold := PInteger(_srcInfBytes)^;
      Inc(LongWord(_srcInfBytes), 4);
      Bits := 31;
    end else
    begin
      Hold := ProcessTail;
      Bits := 7;
    end;
    Code := Tree^[Code + (Hold and 1)];
    Hold := Integer(LongWord(Hold) shr 1);
  until Code <= 0;
CodeFound:
  _holdInf := Hold;
  _bitsInf := Bits;
  Result := -Code;
end;

procedure LoadChLenLengths;
var
  N, M: Integer;
begin
  G_FillLongs(0, _bitLen, ChLenCount);
  G_FillLongs(0, _bitCount, MaxChLenBits + 1);
  N := GetNBits(4);
  M := GetNBits(3);
  _bitLen^[9] := M;
  Inc(_bitCount^[M]);
  M := GetNBits(3);
  _bitLen^[10] := M;
  Inc(_bitCount^[M]);
  M := GetNBits(3);
  _bitLen^[8] := M;
  Inc(_bitCount^[M]);
  M := GetNBits(3);
  _bitLen^[15] := M;
  Inc(_bitCount^[M]);
  M := GetNBits(3);
  _bitLen^[7] := M;
  Inc(_bitCount^[M]);
  if N = 0 then
    Exit;
  M := GetNBits(3);
  _bitLen^[11] := M;
  Inc(_bitCount^[M]);
  if N = 1 then
    Exit;
  M := GetNBits(3);
  _bitLen^[0] := M;
  Inc(_bitCount^[M]);
  if N = 2 then
    Exit;
  M := GetNBits(3);
  _bitLen^[6] := M;
  Inc(_bitCount^[M]);
  if N = 3 then
    Exit;
  M := GetNBits(3);
  _bitLen^[16] := M;
  Inc(_bitCount^[M]);
  if N = 4 then
    Exit;
  M := GetNBits(3);
  _bitLen^[5] := M;
  Inc(_bitCount^[M]);
  if N = 5 then
    Exit;
  M := GetNBits(3);
  _bitLen^[12] := M;
  Inc(_bitCount^[M]);
  if N = 6 then
    Exit;
  M := GetNBits(3);
  _bitLen^[4] := M;
  Inc(_bitCount^[M]);
  if N = 7 then
    Exit;
  M := GetNBits(3);
  _bitLen^[17] := M;
  Inc(_bitCount^[M]);
  if N = 8 then
    Exit;
  M := GetNBits(3);
  _bitLen^[13] := M;
  Inc(_bitCount^[M]);
  if N = 9 then
    Exit;
  M := GetNBits(3);
  _bitLen^[3] := M;
  Inc(_bitCount^[M]);
  if N = 10 then
    Exit;
  M := GetNBits(3);
  _bitLen^[18] := M;
  Inc(_bitCount^[M]);
  if N = 11 then
    Exit;
  M := GetNBits(3);
  _bitLen^[19] := M;
  Inc(_bitCount^[M]);
  if N = 12 then
    Exit;
  M := GetNBits(3);
  _bitLen^[2] := M;
  Inc(_bitCount^[M]);
  if N = 13 then
    Exit;
  M := GetNBits(3);
  _bitLen^[1] := M;
  Inc(_bitCount^[M]);
  if N = 14 then
    Exit;
  M := GetNBits(3);
  _bitLen^[14] := M;
  Inc(_bitCount^[M]);
end;

procedure LoadChLenTree;
label
  NextLoop;
var
  N, M, P, Code, TreeLen, I: Integer;
begin
  LoadChLenLengths();
  G_FillLongs(0, _chLenTree, ChLenTreeSize);
  _nextInfCode^[1] := 0;
  N := _bitCount^[1] shl 1;
  _nextInfCode^[2] := N;
  N := (N + _bitCount^[2]) shl 1;
  _nextInfCode^[3] := N;
  N := (N + _bitCount^[3]) shl 1;
  _nextInfCode^[4] := N;
  N := (N + _bitCount^[4]) shl 1;
  _nextInfCode^[5] := N;
  N := (N + _bitCount^[5]) shl 1;
  _nextInfCode^[6] := N;
  N := (N + _bitCount^[6]) shl 1;
  _nextInfCode^[7] := N;
  TreeLen := 2;
  I := 0;
  while I < ChLenCount do
  begin
    N := _bitLen^[I];
    if N = 0 then
      goto NextLoop;
    M := _nextInfCode^[N];
    Code := Integer(G_ReverseBits(LongWord(M), N));
    _nextInfCode^[N] := M + 1;
    P := 1;
    repeat
      Inc(P, Code and 1);
      Code := Code shr 1;
      Dec(N);
      if N <> 0 then
      begin
        M := P;
        P := _chLenTree^[P];
        if P = 0 then
        begin
          P := TreeLen + 1;
          TreeLen := P + 1;
          _chLenTree^[M] := P;
        end;
      end else
      begin
        _chLenTree^[P] := -I;
        Break;
      end;
    until False;
  NextLoop:
    Inc(I);
  end;
end;

procedure LoadCharDistLengths(Count: Integer);
var
  C, LastLen: Integer;
  P: PInteger;
begin
  LastLen := 7;
  P := PInteger(_bitLen);
  G_FillLongs(0, _bitCount, MaxBits + 1);
  while Count > 0 do
  begin
    C := GetCode(_chLenTree);
    if C < 15 then
    begin
      P^ := C;
      Inc(_bitCount^[C]);
      Inc(LongWord(P), 4);
      LastLen := C;
      Dec(Count);
    end else
    begin
      case C of
        15: C := 2;
        16: C := GetBit + 3;
        17: C := GetNBits(2) + 5;
        18: C := GetNBits(3) + 9;
        19: C := GetNBits(7) + 17;
      end;
      Dec(Count, C);
      Inc(_bitCount^[LastLen], C);
      repeat
        P^ := LastLen;
        Inc(LongWord(P), 4);
        Dec(C);
      until C = 0;
    end;
  end;
end;

procedure LoadCharTree;
label
  NextLoop;
var
  N, M, P, Code, TreeLen, I: Integer;
begin
  G_FillLongs(0, _bitLen, CharCount);
  LoadCharDistLengths(GetNBits(5) + 256);
  G_FillLongs(0, _charTree, CharTreeSize);
  _nextInfCode^[1] := 0;
  N := _bitCount^[1] shl 1;
  _nextInfCode^[2] := N;
  N := (N + _bitCount^[2]) shl 1;
  _nextInfCode^[3] := N;
  N := (N + _bitCount^[3]) shl 1;
  _nextInfCode^[4] := N;
  N := (N + _bitCount^[4]) shl 1;
  _nextInfCode^[5] := N;
  N := (N + _bitCount^[5]) shl 1;
  _nextInfCode^[6] := N;
  N := (N + _bitCount^[6]) shl 1;
  _nextInfCode^[7] := N;
  N := (N + _bitCount^[7]) shl 1;
  _nextInfCode^[8] := N;
  N := (N + _bitCount^[8]) shl 1;
  _nextInfCode^[9] := N;
  N := (N + _bitCount^[9]) shl 1;
  _nextInfCode^[10] := N;
  N := (N + _bitCount^[10]) shl 1;
  _nextInfCode^[11] := N;
  N := (N + _bitCount^[11]) shl 1;
  _nextInfCode^[12] := N;
  N := (N + _bitCount^[12]) shl 1;
  _nextInfCode^[13] := N;
  N := (N + _bitCount^[13]) shl 1;
  _nextInfCode^[14] := N;
  TreeLen := 2;
  I := 0;
  while I < CharCount do
  begin
    N := _bitLen^[I];
    if N = 0 then
      goto NextLoop;
    M := _nextInfCode^[N];
    Code := Integer(G_ReverseBits(LongWord(M), N));
    _nextInfCode^[N] := M + 1;
    P := 1;
    repeat
      Inc(P, Code and 1);
      Code := Code shr 1;
      Dec(N);
      if N <> 0 then
      begin
        M := P;
        P := _charTree^[P];
        if P = 0 then
        begin
          P := TreeLen + 1;
          TreeLen := P + 1;
          _charTree^[M] := P;
        end;
      end else
      begin
        _charTree^[P] := -I;
        Break;
      end;
    until False;
  NextLoop:
    Inc(I);
  end;
end;

procedure LoadDistTree;
label
  NextLoop;
var
  N, M, P, Code, TreeLen, I: Integer;
begin
  G_FillLongs(0, _bitLen, DistCount);
  LoadCharDistLengths(GetNBits(6) + 1);
  G_FillLongs(0, _distTree, DistTreeSize);
  _nextInfCode^[1] := 0;
  N := _bitCount^[1] shl 1;
  _nextInfCode^[2] := N;
  N := (N + _bitCount^[2]) shl 1;
  _nextInfCode^[3] := N;
  N := (N + _bitCount^[3]) shl 1;
  _nextInfCode^[4] := N;
  N := (N + _bitCount^[4]) shl 1;
  _nextInfCode^[5] := N;
  N := (N + _bitCount^[5]) shl 1;
  _nextInfCode^[6] := N;
  N := (N + _bitCount^[6]) shl 1;
  _nextInfCode^[7] := N;
  N := (N + _bitCount^[7]) shl 1;
  _nextInfCode^[8] := N;
  N := (N + _bitCount^[8]) shl 1;
  _nextInfCode^[9] := N;
  N := (N + _bitCount^[9]) shl 1;
  _nextInfCode^[10] := N;
  N := (N + _bitCount^[10]) shl 1;
  _nextInfCode^[11] := N;
  N := (N + _bitCount^[11]) shl 1;
  _nextInfCode^[12] := N;
  N := (N + _bitCount^[12]) shl 1;
  _nextInfCode^[13] := N;
  N := (N + _bitCount^[13]) shl 1;
  _nextInfCode^[14] := N;
  TreeLen := 2;
  I := 0;
  while I < DistCount do
  begin
    N := _bitLen^[I];
    if N = 0 then
      goto NextLoop;
    M := _nextInfCode^[N];
    Code := Integer(G_ReverseBits(LongWord(M), N));
    _nextInfCode^[N] := M + 1;
    P := 1;
    repeat
      Inc(P, Code and 1);
      Code := Code shr 1;
      Dec(N);
      if N <> 0 then
      begin
        M := P;
        P := _distTree^[P];
        if P = 0 then
        begin
          P := TreeLen + 1;
          TreeLen := P + 1;
          _distTree^[M] := P;
        end;
      end else
      begin
        _distTree^[P] := -I;
        Break;
      end;
    until False;
  NextLoop:
    Inc(I);
  end;
end;

procedure LoadFixedTree;
var
  I, N, M, P, Code, TreeLen: Integer;
begin
  G_FillLongs(0, _charTree, CharTreeSize);
  _nextInfCode^[1] := 0;
  N := FixedBitCount[1] shl 1;
  _nextInfCode^[2] := N;
  I := 0;
  while I < MaxBits do
  begin
    N := (N + FixedBitCount[I]) shl 1;
    _nextInfCode^[I + 1] := N;
    Inc(I);
  end;
  TreeLen := 2;
  I := 0;
  while I < CharCount do
  begin
    N := FixedCharLength[I];
    M := _nextInfCode^[N];
    Code := Integer(G_ReverseBits(LongWord(M), N));
    _nextInfCode^[N] := M + 1;
    P := 1;
    repeat
      Inc(P, Code and 1);
      Code := Code shr 1;
      Dec(N);
      if n <> 0 then
      begin
        M := P;
        P := _charTree^[P];
        if P = 0 then
        begin
          P := TreeLen + 1;
          TreeLen := P + 1;
          _charTree^[M] := P;
        end;
      end else
      begin
        _charTree^[P] := -I;
        Break;
      end;
    until False;
    Inc(I);
  end;
end;

procedure ReadNonCompressedBlock;
var
  Bits, Hold, Counter: Integer;
  P: PWord;
begin
  Counter := GetNBits(8) + BlockSize;
  if Counter > _outCounter then
    Counter := _outCounter;
  Dec(_outCounter, Counter);
  Bits := _bitsInf;
  Hold := _holdInf;
  P := PWord(_srcInfBytes);
  while Counter > 0 do
  begin
    if Bits < 8 then
      if _srcIndex < _break16Offset then
      begin
        Inc(_srcIndex, 2);
        Hold := Hold or (P^ shl Bits);
        Inc(Bits, 16);
        Inc(LongWord(P), 2);
      end else
      begin
        _srcInfBytes := PBytes(P);
        Hold := Hold or ProcessNBitsTail(Bits);
        P := PWord(_srcInfBytes);
        Inc(Bits, 8);
      end;
    PByte(_dstInfBytes)^ := Byte(Hold);
    Hold := Hold shr 8;
    Dec(Bits, 8);
    Inc(LongWord(_dstInfBytes));
    Dec(Counter);
  end;
  _srcInfBytes := PBytes(P);
  _bitsInf := Bits;
  _holdInf := Hold;
end;

procedure InflateCore;
var
  C, Length: Integer;
begin
  repeat
    _fixedTree := False;
    if GetBit <> 0 then
    begin
      LoadChLenTree;
      LoadCharTree;
      LoadDistTree;
    end
    else if GetBit <> 0 then
    begin
      LoadFixedTree;
      _fixedTree := True;
    end else
    begin
      ReadNonCompressedBlock;
      if _outCounter > 0 then
        Continue
      else
        Exit;
    end;
    _inCounter := BlockSize;
    repeat
      C := GetCode(_charTree);
      Dec(_inCounter);
      if C < FirstLengthChar then
      begin
        PByte(_dstInfBytes)^ := Byte(C);
        Inc(LongWord(_dstInfBytes));
        Dec(_outCounter);
      end else
      begin
        Dec(C, FirstCharWithExBit);
        if C < 0 then
          Length := C + 17
        else if C < 16 then
          Length := GetNBits(CharExBitLength[C]) + CharExBitBase[C]
        else
        begin
          C := GetNBits(8);
          Length := C and $7F;
          if (C and $80) = 0 then
          begin
            C := GetNBits(8);
            Length := Length or ((C and $7F) shl 7);
            if (C and $80) = 0 then
            begin
              C := GetNBits(7);
              Length := Length or (C shl 14);
            end;
          end;
          Inc(Length, 145);
        end;
        if not _fixedTree then
          C := GetCode(_distTree)
        else
          C := GetNBits(6);
        if C >= FirstDistWithExBit then
          C := GetNBits(DistExBitLength[C]) + DistExBitBase[C]
        else
          C := DistExBitBase[C];
        G_CopyBytes(@_dstInfBytes^[-C], _dstInfBytes, Length);
        Inc(LongWord(_dstInfBytes), Length);
        Dec(_outCounter, Length);
      end;
      if _outCounter <= 0 then
        Exit;
    until _inCounter = 0;
  until False;
end;

function G_GetInflatedLength(SourceBytes: Pointer): Integer;
begin
  if not Assigned(SourceBytes) then
    RaiseErrorFmt(SErrArgumentNull, 'SourceBytes', 'G_GetInflatedLength');
  Result := PInteger(SourceBytes)^;
  if Result < 0 then
    Result := -Result;
end;

function G_Inflate(SourceBytes: Pointer; SourceLength: Integer;
  var DestinationBytes: Pointer; BeforeGap, AfterGap: Integer): Integer;
var
  SkipInflating: Boolean;
begin
  if not Assigned(SourceBytes) then
    RaiseErrorFmt(SErrArgumentNull, 'SourceBytes', 'G_Inflate');
  Result := PInteger(SourceBytes)^;
  SkipInflating := False;
  if Result < 0 then
  begin
    Result := -Result;
    SkipInflating := True;
  end;
  GetMem(DestinationBytes, BeforeGap + Result + AfterGap);
  if BeforeGap > 0 then
    G_ZeroMem(DestinationBytes, BeforeGap);
  if SkipInflating then
    G_CopyMem(@PBytes(SourceBytes)^[4], @PBytes(DestinationBytes)^[BeforeGap], Result)
  else if Result > 0 then
    G_Inflate(SourceBytes, SourceLength, @PBytes(DestinationBytes)^[BeforeGap]);
  if AfterGap > 0 then
    G_ZeroMem(@PBytes(DestinationBytes)^[BeforeGap + Result], AfterGap);
end;

var
  InflatorCriticalSection: _RTL_CRITICAL_SECTION;

procedure AllocInflator;
begin
  if _bitLen = nil then
  begin
    GetMem(_charTree, CharTreeSize * SizeOf(Integer));
    GetMem(_distTree, DistTreeSize * SizeOf(Integer));
    GetMem(_chLenTree, ChLenTreeSize * SizeOf(Integer));
    GetMem(_bitCount, (MaxBits + 1) * SizeOf(Integer));
    GetMem(_nextInfCode, (MaxBits + 1) * SizeOf(Integer));
    GetMem(_bitLen, CharCount * SizeOf(Integer));
  end;
end;

function G_Inflate(SourceBytes: Pointer; SourceLength: Integer;
  DestinationBytes: Pointer): Integer;
begin
  if not Assigned(SourceBytes) then
    RaiseErrorFmt(SErrArgumentNull, 'SourceBytes', 'G_Inflate');
  Result := PInteger(SourceBytes)^;
  if (Result = 0) or not Assigned(DestinationBytes) then
  begin
    if Result < 0 then
      Result := -Result;
    Exit;
  end;
  if Result < 0 then
  begin
    Result := -Result;
    G_CopyMem(@PBytes(SourceBytes)^[4], DestinationBytes, Result);
    Exit;
  end;
  if IsMultiThread then
    EnterCriticalSection(InflatorCriticalSection);
  try
    AllocInflator;
    _srcInfBytes := @PBytes(SourceBytes)^[4];
    _dstInfBytes := PBytes(DestinationBytes);
    _bitsInf := 0;
    _holdInf := 0;
    _srcIndex := 4;
    _breakInfOffset := SourceLength;
    _break32Offset := SourceLength - 3;
    _break16Offset := SourceLength - 1;
    _outCounter := Result;
    InflateCore;
  finally
    if IsMultiThread then
      LeaveCriticalSection(InflatorCriticalSection);
  end;
end;

procedure G_ReleaseInflator;
begin
  if IsMultiThread then
    EnterCriticalSection(InflatorCriticalSection);
  try
    if _bitLen <> nil then
    begin
      FreeMem(_charTree);
      FreeMem(_distTree);
      FreeMem(_chLenTree);
      FreeMem(_bitCount);
      FreeMem(_nextInfCode);
      FreeMem(_bitLen);
      _bitLen := nil;
    end;
  finally
    if IsMultiThread then
      LeaveCriticalSection(InflatorCriticalSection);
  end;
end;

initialization
  InitializeCriticalSection(DeflatorCriticalSection);
  InitializeCriticalSection(InflatorCriticalSection);

finalization
  G_ReleaseDeflator;
  G_ReleaseInflator;

  DeleteCriticalSection(DeflatorCriticalSection);
  DeleteCriticalSection(InflatorCriticalSection);

end.

