
/////////////////////////////////////////////////////
//                                                 //
//   AcedLists 1.03                                //
//                                                 //
//        .   //
//                                                 //
//   mailto: acedutils@yandex.ru                   //
//                                                 //
/////////////////////////////////////////////////////

unit AcedLists;

interface

uses Registry, AcedConsts, AcedStreams, AcedCrypto, AcedCompression;

type

{  TBitList -     Boolean. }

  TBitList = class(TObject)
  private
    FCapacity: Integer;
    FItems: PLongWordItemList;
    FCount: Integer;
    procedure SetCount(NewCount: Integer);
    function GetItem(Index: Integer): Boolean;
    procedure SetItem(Index: Integer; V: Boolean);
  public

  {  Create      .
          False. }

    constructor Create(Count: Integer);

  {  Destroy  ,    TBitList
      ,   InnerList. }

    destructor Destroy; override;

  {  }

  {  Items       
      Index (  ). }

    property Items[Index: Integer]: Boolean read GetItem write SetItem; default;

  {  Capacity   (  )  ,
      InnerList. }

    property Capacity: Integer read FCapacity;

  {  Count       
    .        
      InnerList. }

    property Count: Integer read FCount write SetCount;

  {  InnerList     ,  
      Boolean, ..  .    
      ,    Capacity. }

    property InnerList: PLongWordItemList read FItems;

  {  }

  {   Load      Reader. }

    procedure Load(Reader: TBinaryReader); overload;

  {   Load    ,  
    BitList,   . }

    procedure Load(BitList: TBitList); overload;

  {   Load        
    ,   Items.  Count   
     ( ).      TBitList.Count
      . }

    procedure Load(Items: Pointer; Count: Integer); overload;

  {   Load    ,   
    Windows.    TRegistry,   Registry,
          ,  .  
      ValueName.   EncryptionKey   nil,
    ,         
        EncryptionKey. }

    procedure Load(Registry: TRegistry; const ValueName: string;
      EncryptionKey: PSHA256Digest = nil); overload;

  {   Save      Writer. }

    procedure Save(Writer: TBinaryWriter); overload;

  {   Save     Windows.  
     TRegistry,   Registry,   
       ,  .    
    ValueName.   EncryptionKey   nil,  
        EncryptionKey.  CompressionMode,
        dcmNoCompression,    . }

    procedure Save(Registry: TRegistry; const ValueName: string;
      EncryptionKey: PSHA256Digest = nil;
      CompressionMode: TCompressionMode = dcmNoCompression); overload;

  {  FromBase64     S,  
          Base64.  , 
      , ,     
    EncryptionKey. }

    procedure FromBase64(const S: string; EncryptionKey: PSHA256Digest = nil);

  {  ToBase64        
        Base64.     
    .   Encrypted   nil, ,  
     ,    EncryptionKey.  
    CompressionMode,     dcmNoCompression,  
     . }

    function ToBase64(EncryptionKey: PSHA256Digest = nil;
      CompressionMode: TCompressionMode = dcmNoCompression): string;

  {  Invert       Index
    (  ).     True,   False, 
      False,   True. }

    function Invert(Index: Integer): Boolean;

  {  IndexOf     , 
      V.       StartIndex.
      V    ,   -1. }

    function IndexOf(V: Boolean; StartIndex: Integer = 0): Integer;

  {  LastIndexOf     ,
       V.     
        StartIndex, .   V
       ,   -1. }

    function LastIndexOf(V: Boolean; StartIndex: Integer = MaxInt): Integer;

  {  CountOf     , 
      V. }

    function CountOf(V: Boolean): Integer;

  {  SetAll  , .. Count    
    True  False      V.  
        TBitList. }

    function SetAll(V: Boolean): TBitList;

  {  NotSet    . ,  
     False,  True,     True,  False.
          TBitList. }

    function NotSet: TBitList;

  {  AndSet      
      BitList.  Count    .
           TBitList,  
      . }

    function AndSet(BitList: TBitList): TBitList;

  {  OrSet      
      BitList.  Count    .
           TBitList,  
      . }

    function OrSet(BitList: TBitList): TBitList;

  {  XorSet   " "   
          BitList.  Count
       .     
      TBitList,     . }

    function XorSet(BitList: TBitList): TBitList;

  {  AndNotSet   "-"    
         BitList.   
    ,    True    , 
      True     False   BitList.  Count
       .     
      TBitList,     . }

    function AndNotSet(BitList: TBitList): TBitList;

  {  Equals  True,     
      BitList.      
       ,   False. }

    function Equals(BitList: TBitList): Boolean;

  {  Clone    TBitList,  
      . }

    function Clone: TBitList;
  end;

  
{  TIntegerList -     Integer. }

  TIntegerList = class(TObject)
  private
    FCapacity: Integer;
    FItems: PIntegerItemList;
    FCount: Integer;
    FMaintainSorted: Boolean;
    FSorted: Boolean;
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetMaintainSorted(Value: Boolean);
  public

  {   Create    TIntegerList. 
         .    
        .  Sorted  MaintainSorted
       True. }

    constructor Create; overload;

  {   Create    TIntegerList 
       InitialCapacity    .
     Sorted  MaintainSorted    True. }

    constructor Create(InitialCapacity: Integer); overload;

  {  Destroy  ,   
    TIntegerList    ,   ItemList. }

    destructor Destroy; override;

  {  }

  {  Capacity     ,  
        .    
         Capacity, 
       Capacity   , 
     G_EnlargeCapacity(Capacity). }

    property Capacity: Integer read FCapacity write SetCapacity;

  {  Count      
     .       
     . }

    property Count: Integer read FCount write FCount;

  {  ItemList     , 
       .      
    Capacity.    ItemList    
      . }

    property ItemList: PIntegerItemList read FItems;

  {  MaintainSorted    , ,
           .  
      True,    Insert  
      . }

    property MaintainSorted: Boolean read FMaintainSorted write SetMaintainSorted;

  {   Sorted  True,     ,
     False -   .     ,  
    ,   EnsureSorted. }

    property Sorted: Boolean read FSorted;

  {  }

  {   Load      Reader. }

    procedure Load(Reader: TBinaryReader); overload;

  {   Load    ,  
    IntegerList. }

    procedure Load(IntegerList: TIntegerList); overload;

  {   Load       Integer,
      Items,   Count . 
    Sorted ,      . }

    procedure Load(Items: Pointer; Count: Integer; Sorted: Boolean); overload;

  {   Load     Windows.  
     TRegistry,   Registry,   
       ,  .    
    ValueName.   EncryptionKey   nil, , 
             
       EncryptionKey. }

    procedure Load(Registry: TRegistry; const ValueName: string;
      EncryptionKey: PSHA256Digest = nil); overload;

  {   Save      Writer. }

    procedure Save(Writer: TBinaryWriter); overload;

  {   Save     Windows.  
     TRegistry,   Registry,   
       ,  .    
    ValueName.   EncryptionKey   nil,  
        EncryptionKey.  CompressionMode,
        dcmNoCompression,    . }

    procedure Save(Registry: TRegistry; const ValueName: string;
      EncryptionKey: PSHA256Digest = nil;
      CompressionMode: TCompressionMode = dcmNoCompression); overload;

  {  FromBase64     S,   
         Base64.  ,  
     , ,     
    EncryptionKey. }

    procedure FromBase64(const S: string; EncryptionKey: PSHA256Digest = nil);

  {  ToBase64        
        Base64.     
    .   EncryptionKey   nil, , 
      ,    EncryptionKey. 
    CompressionMode,     dcmNoCompression,  
     . }

    function ToBase64(EncryptionKey: PSHA256Digest = nil;
      CompressionMode: TCompressionMode = dcmNoCompression): string;

  {   EnsureCapacity ,    ,
      ItemList,     Capacity
    , ..  Capacity    TIntegerList
        ,   Capacity.}

    procedure EnsureCapacity(Capacity: Integer);

  {  Add   N   .  
    MaintainSorted  True,     
    ,  -   . }

    procedure Add(N: Integer);

  {  AddIfNotExists   N     ,
         .   MaintainSorted  True,
         ,  -   . }

    procedure AddIfNotExists(N: Integer);

  {  Insert     N   Index 
    (  ).       
    MaintainSorted  False. }

    procedure Insert(Index, N: Integer);

  {  Peek     , .. 
      (Count - 1). }

    function Peek: Integer;

  {  Pop       
       .     
     .         Add,
      MaintainSorted   False. }

    function Pop: Integer;

  {  IndexOf       N  -1,
         . }

    function IndexOf(N: Integer): Integer;

  {  Remove       N,   
      . }

    procedure Remove(N: Integer);

  {  RemoveAt    ,    Index
    (  ). }

    procedure RemoveAt(Index: Integer);

  {  UnorderedRemove       N,  
       .     
    ,    Count   1.  
          MaintainSorted  False.
       ,  Remove,    
       . }

    procedure UnorderedRemove(N: Integer);

  {  UnorderedRemoveAt    ,   
    Index (  ).     
    ,    Count   1.  
          MaintainSorted  False.
       ,  RemoveAt,    
       . }

    procedure UnorderedRemoveAt(Index: Integer);

  {  Clear       Count. ,
       ,      
     Capacity  . }

    procedure Clear;

  {  EnsureSorted   ,  ,  
       . }

    procedure EnsureSorted;

  {  TrimToSize    ,  
    ItemList,  ,      
      . }

    procedure TrimToSize;

  {  Equals  True,     
      IntegerList.      
       ,   False. }

    function Equals(IntegerList: TIntegerList): Boolean;

  {  Clone    TIntegerList,  
      . }

    function Clone: TIntegerList;
  end;

  
{  TWordList -     Word. }

  TWordList = class(TObject)
  private
    FCapacity: Integer;
    FItems: PWordItemList;
    FCount: Integer;
    FMaintainSorted: Boolean;
    FSorted: Boolean;
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetMaintainSorted(Value: Boolean);
  public

  {   Create c   TWordList. 
         .    
        .  Sorted  MaintainSorted
       True. }

    constructor Create; overload;

  {   Create    TWordList  
      InitialCapacity    .  Sorted
     MaintainSorted    True. }

    constructor Create(InitialCapacity: Integer); overload;

  {  Destroy  ,    TWordList
      ,   ItemList. }

    destructor Destroy; override;

  {  }

  {  Capacity     ,  
        .    
         Capacity, 
       Capacity   , 
     G_EnlargeCapacity(Capacity). }

    property Capacity: Integer read FCapacity write SetCapacity;

  {  Count      
     .       
     . }

    property Count: Integer read FCount write FCount;

  {  ItemList     , 
       .      
    Capacity.    ItemList    
      . }

    property ItemList: PWordItemList read FItems;

  {  MaintainSorted    , ,
           .  
      True,    Insert  
      . }

    property MaintainSorted: Boolean read FMaintainSorted write SetMaintainSorted;

  {   Sorted  True,     ,
     False -   .     ,  
    ,   EnsureSorted. }

    property Sorted: Boolean read FSorted;

  {  }

  {   Load      Reader. }

    procedure Load(Reader: TBinaryReader); overload;

  {   Load    ,  
    WordList. }

    procedure Load(WordList: TWordList); overload;

  {   Load       Word,
      Items,   Count . 
    Sorted ,      . }

    procedure Load(Items: Pointer; Count: Integer; Sorted: Boolean); overload;

  {   Load     Windows.  
     TRegistry,   Registry,   
       ,  .    
    ValueName.   EncryptionKey   nil, , 
             
       EncryptionKey. }

    procedure Load(Registry: TRegistry; const ValueName: string;
      EncryptionKey: PSHA256Digest = nil); overload;

  {   Save      Writer. }

    procedure Save(Writer: TBinaryWriter); overload;

  {   Save     Windows.  
     TRegistry,   Registry,   
       ,  .    
    ValueName.   EncryptionKey   nil,   
        EncryptionKey.  CompressionMode,
        dcmNoCompression,    . }

    procedure Save(Registry: TRegistry; const ValueName: string;
      EncryptionKey: PSHA256Digest = nil;
      CompressionMode: TCompressionMode = dcmNoCompression); overload;

  {  FromBase64     S,   
         Base64.  ,  
     , ,     
    EncryptionKey. }

    procedure FromBase64(const S: string; EncryptionKey: PSHA256Digest = nil);

  {  ToBase64        
        Base64.     
    .   EncryptionKey   nil, , 
      ,    EncryptionKey. 
    CompressionMode,     dcmNoCompression,  
     . }

    function ToBase64(EncryptionKey: PSHA256Digest = nil;
      CompressionMode: TCompressionMode = dcmNoCompression): string;

  {   EnsureCapacity ,    ,
      ItemList,     Capacity
    , ..  Capacity    TWordList
        ,   Capacity. }

    procedure EnsureCapacity(Capacity: Integer);

  {  Add   N   .  
    MaintainSorted  True,     
    ,  -   . }

    procedure Add(N: Word);

  {  AddIfNotExists   N     ,
         .   MaintainSorted  True,
         ,  -   . }

    procedure AddIfNotExists(N: Word);

  {  Insert     N   Index 
    (  ).       
    MaintainSorted  False. }

    procedure Insert(Index: Integer; N: Word);

  {  Peek     , .. 
      (Count - 1). }

    function Peek: Word;

  {  Pop       
       .     
     .         Add,
      MaintainSorted   False. }

    function Pop: Word;

  {  IndexOf       N  -1,
         . }

    function IndexOf(N: Word): Integer;

  {  Remove       N,   
      . }

    procedure Remove(N: Word);

  {  RemoveAt    ,    Index
    (  ). }

    procedure RemoveAt(Index: Integer);

  {  UnorderedRemove       N,  
       .     
    ,    Count   1.  
          MaintainSorted  False.
       ,  Remove,    
       . }

    procedure UnorderedRemove(N: Word);

  {  UnorderedRemoveAt    ,   
    Index (  ).     
    ,    Count   1.  
          MaintainSorted  False.
       ,  RemoveAt,    
       . }

    procedure UnorderedRemoveAt(Index: Integer);

  {  Clear       Count. ,
       ,      
     Capacity  . }

    procedure Clear;

  {  EnsureSorted   ,  ,  
       . }

    procedure EnsureSorted;

  {  TrimToSize    ,  
    ItemList,  ,      
      . }

    procedure TrimToSize;

  {  Equals  True,     
      WordList.      
       ,   False. }

    function Equals(WordList: TWordList): Boolean;

  {  Clone    TWordList,  
      . }

    function Clone: TWordList;
  end;

  
{  TIntegerAssociatedList    ,
        Integer    TObject. 
      . }

  TIntegerAssociatedList = class(TObject)
  private
    FCapacity: Integer;
    FKeys: PIntegerItemList;
    FValues: PObjectItemList;
    FCount: Integer;
    FOwnValues: Boolean;
    procedure SetCapacity(NewCapacity: Integer);
    function GetItem(Key: Integer): TObject;
    procedure SetItem(Key: Integer; Value: TObject);
  public

  {   Create   
    TIntegerAssociatedList.       
      .      
       . }

    constructor Create; overload;

  {   Create   
    TIntegerAssociatedList     InitialCapacity 
          . }

    constructor Create(InitialCapacity: Integer); overload;

  {  Destroy  ,   
        TIntegerAssociatedList.   OwnValues
     True,    Free    
     . }

    destructor Destroy; override;

  {  }

  {  Items    , 
      Key.  ,      , 
     nil.  ,    ,  
      -.      ,
      ,    .  ,
      OwnValues  True,    
     Free. }

    property Items[Key: Integer]: TObject read GetItem write SetItem; default;

  {  Capacity     ,  
        .    
         Capacity, 
       Capacity   , 
     G_EnlargeCapacity(Capacity). }

    property Capacity: Integer read FCapacity write SetCapacity;

  {  Count      
     .       
     . }

    property Count: Integer read FCount write FCount;

  {   OwnValues ,     Free 
    ,    ,    
    ,        
     .    OwnValues  False   Free
     . }

    property OwnValues: Boolean read FOwnValues write FOwnValues;

  {  KeyList     ,  
     .       Capacity. }

    property KeyList: PIntegerItemList read FKeys;

  {  ValueList      ,
      .      
    Capacity. }

    property ValueList: PObjectItemList read FValues;

  {  }

  {   EnsureCapacity ,    ,
      KeyList  ValueList,   
    Capacity , ..  Capacity   
    TIntegerAssociatedList     , 
     Capacity. }

    procedure EnsureCapacity(Capacity: Integer);

  {  Add      Key  
      Value.         ,
     . }

    procedure Add(Key: Integer; Value: TObject);

  {  IndexOf    Key   ,
      KeyList.      
     ,   -1. }

    function IndexOf(Key: Integer): Integer;

  {  Remove        Key,
       .   OwnValues  True, 
    ,    ,   Free. }

    procedure Remove(Key: Integer);

  {  RemoveAt   -,    Index
      (  ).   OwnValues 
    True,       Free. }

    procedure RemoveAt(Index: Integer);

  {  Clear     Count.   OwnValues
     True,   Free     
    . ,    ,    
     Capacity  . }

    procedure Clear;

  {  TrimToSize    , 
     KeyList  ValueList,  ,   
         . }

    procedure TrimToSize;
  end;

  
{  TIntegerHashTable     (),
       Integer    TObject.  
     . }

  TIntegerHashTable = class(TObject)
  private
    FCapacity: Integer;
    FKeys: PIntegerItemList;
    FValues: PObjectItemList;
    FBits: PLongWordItemList;
    FCount: Integer;
    FMaxCount: Integer;
    FOwnValues: Boolean;
    procedure SetMaxCount(NewMaxCount: Integer);
    function GetIndex(Key: Integer): Integer;
    procedure SetCapacity(NewCapacity: Integer);
    function GetItem(Key: Integer): TObject;
    procedure SetItem(Key: Integer; Value: TObject);
  public

  {   Create    TIntegerHashedList
         ,    10 
     . }

    constructor Create; overload;

  {   Create    TIntegerHashedList
      ,    InitialCapacity 
            . }

    constructor Create(InitialCapacity: Integer); overload;

  {  Destroy  ,   
        TIntegerHashedList.   OwnValues
     True,    Free   
      . }

    destructor Destroy; override;

  {  }

  {  Items    ,  
     Key.  ,      , 
     nil.  ,    ,  
      -.      ,
      ,    .  ,
      OwnValues  True,    
     Free.}

    property Items[Key: Integer]: TObject read GetItem write SetItem; default;

  {  Capacity     , 
        .    
         Capacity, 
         ,   
        . }

    property Capacity: Integer read FMaxCount write SetMaxCount;

  {  Count      
    . }

    property Count: Integer read FCount;

  {   OwnValues ,     Free
     ,    ,   
     ,        
     .    OwnValues  False   Free
     . }

    property OwnValues: Boolean read FOwnValues write FOwnValues;

  {  InnerKeyList     , 
          .    
       InnerCapacity.  , 
     ,      .
       , .. , 
     ,  ,   , 
     InnerBits. }

    property InnerKeyList: PIntegerItemList read FKeys;

  {  InnerValueList      ,
      .      
    InnerCapacity. }

    property InnerValueList: PObjectItemList read FValues;

  {  InnerBits     , 
        InnerKeyList  InnerValueList. ,
      1, ,     InnerKeyList
         ,    
    InnerValueList     - ,   
    .    0,   
     . }

    property InnerBits: PLongWordItemList read FBits;

  {  InnerCapacity   ,  
        InnerKeyList, InnerValueList,  
        InnerBits.    
     7/5    Capacity, ..  
             
      71-73% . }

    property InnerCapacity: Integer read FCapacity;

  {  }

  {   EnsureCapacity ,    ,
         Capacity , .. 
    Capacity    TIntegerHashedList   
    EnsureCapacity     ,  
    Capacity. }

    procedure EnsureCapacity(Capacity: Integer);

  {  Add      Key  
      Value.         ,
     . }

    procedure Add(Key: Integer; Value: TObject);

  {  Contains  True,     
       Key.        ,
      False. }

    function Contains(Key: Integer): Boolean;

  {  Remove        Key, 
      .   OwnValues  True,  ,
       ,   Free. }

    procedure Remove(Key: Integer);

  {  Clear   InnerBits     
    Count.    OwnValues  True,   Free
         . ,  
     ,     Capacity  . }

    procedure Clear;
  end;

  
{  TStringAssociatedList    ,
        String    TObject. 
      . }

  TStringAssociatedList = class(TObject)
  private
    FCapacity: Integer;
    FKeys: PStringItemList;
    FValues: PObjectItemList;
    FCount: Integer;
    FCaseSensitive: Boolean;
    FOwnValues: Boolean;
    procedure SetCapacity(NewCapacity: Integer);
    function GetItem(const Key: string): TObject;
    function SearchFirstGE(const Key: string): Integer;
    procedure SetItem(const Key: string; Value: TObject);
  public

  {   Create    TStringAssociatedList.
             .
            
    .  CaseSensitive ,    
           String.   
     False,         ,
     True, . }

    constructor Create(CaseSensitive: Boolean = False); overload;

  {   Create    TStringAssociatedList
        InitialCapacity    
         .  CaseSensitive
    ,         
      String.     False,   
         ,  True,   }

    constructor Create(CaseSensitive: Boolean; InitialCapacity: Integer); overload;

  {  Destroy  ,    
       TStringAssociatedList.   OwnValues
     True,    Free   
      . }

    destructor Destroy; override;

  {  }

  {  Items    , 
      Key.  ,      , 
     nil.  ,    ,  
      -.      ,
      ,    .  ,
      OwnValues  True,    
     Free. }

    property Items[const Key: string]: TObject read GetItem write SetItem; default;

  {  CaseSensitive    ,
        .     True,
       , ..   ,   
     ,  False     
    . }

    property CaseSensitive: Boolean read FCaseSensitive;

  {  Capacity     ,  
        .    
         Capacity, 
       Capacity   , 
     G_EnlargeCapacity(Capacity). }

    property Capacity: Integer read FCapacity write SetCapacity;

  {  Count      
     .       
     . }

    property Count: Integer read FCount write FCount;

  {   OwnValues ,     Free
     ,    ,   
     ,        
     .    OwnValues  False   Free
     . }

    property OwnValues: Boolean read FOwnValues write FOwnValues;

  {  KeyList     , 
      .       Capacity. }

    property KeyList: PStringItemList read FKeys;

  {  ValueList      ,
      .      
    Capacity. }

    property ValueList: PObjectItemList read FValues;

  {  }

  {   EnsureCapacity ,    ,
      KeyList  ValueList,   
    Capacity , ..  Capacity   
    TStringAssociatedList     , 
     Capacity. }

    procedure EnsureCapacity(Capacity: Integer);

  {  Add      Key  
      Value.         ,
     . }

    procedure Add(const Key: string; Value: TObject);

  {  IndexOf    Key   ,
      KeyList.      
     ,   -1. }

    function IndexOf(const Key: string): Integer;

  {  Remove        Key,
       .   OwnValues  True, 
    ,    ,   Free. }

    procedure Remove(const Key: string);

  {  RemoveAt   -,    Index
      (  ).   OwnValues 
    True,       Free. }

    procedure RemoveAt(Index: Integer);

  {  Clear     Count.   OwnValues
     True,   Free     
    . ,    ,    
     Capacity  . }

    procedure Clear;

  {  TrimToSize    ,  
    KeyList  ValueList,  ,    
        . }

    procedure TrimToSize;
  end;

  
{  TStringHashTable     (), 
      String    TObject.    
   . }

  TStringHashTable = class(TObject)
  private
    FCapacity: Integer;
    FKeys: PStringItemList;
    FValues: PObjectItemList;
    FCount: Integer;
    FMaxCount: Integer;
    FCaseSensitive: Boolean;
    FOwnValues: Boolean;
    procedure SetMaxCount(NewMaxCount: Integer);
    function GetIndex(P: Pointer): Integer;
    procedure SetCapacity(NewCapacity: Integer);
    function GetItem(const Key: string): TObject;
    procedure SetItem(const Key: string; Value: TObject);
  public

  {   Create    TStringHashedList
         ,    10 
     .  CaseSensitive ,   
            String.  
      False,       
     ,  True, . }

    constructor Create(CaseSensitive: Boolean = False); overload;

  {   Create    TStringHashedList
      ,    InitialCapacity 
            . 
    CaseSensitive ,      
         String.     False,
            ,  True,
     . }

    constructor Create(CaseSensitive: Boolean; InitialCapacity: Integer); overload;

  {  Destroy  ,   
        TStringHashedList.   OwnValues
     True,    Free   
      . }

    destructor Destroy; override;

  {  }

  {  Items    , 
      Key.  ,      , 
     nil.  ,    ,  
      -.      ,
      ,    .  ,
      OwnValues  True,    
     Free. }

    property Items[const Key: string]: TObject read GetItem write SetItem; default;

  {  CaseSensitive    ,
        .     True,
       , ..   ,   
     ,  False,     
    . }

    property CaseSensitive: Boolean read FCaseSensitive;

  {  Capacity     , 
        .    
         Capacity, 
         ,   
        . }

    property Capacity: Integer read FMaxCount write SetMaxCount;

  {  Count      
    . }

    property Count: Integer read FCount;

  {   OwnValues ,     Free
     ,    ,   
     ,        
     .    OwnValues  False   Free
     . }

    property OwnValues: Boolean read FOwnValues write FOwnValues;

  {  InnerKeyList     , 
          .    
       InnerCapacity.  , 
     ,      ,  nil,
       .     
    , .. PLongWord(@InnerKeyList^[I])^,  $FFFFFFFF. }

    property InnerKeyList: PStringItemList read FKeys;

  {  InnerValueList      ,
      .      
    InnerCapacity. }

    property InnerValueList: PObjectItemList read FValues;

  {  InnerCapacity   ,   
       InnerKeyList  InnerValueList.   
      7/5    Capacity, .. 
             
       71-73% . }

    property InnerCapacity: Integer read FCapacity;

  {  }

  {   EnsureCapacity ,    ,
         Capacity , .. 
    Capacity    TStringHashedList   
    EnsureCapacity     ,  
    Capacity. }

    procedure EnsureCapacity(Capacity: Integer);

  {  Add      Key  
      Value.         ,
     . }

    procedure Add(const Key: string; Value: TObject);

  {  Contains  True,     
       Key.        ,
      False. }

    function Contains(const Key: string): Boolean;

  {  Remove        Key, 
      .   OwnValues  True,  ,
       ,   Free. }

    procedure Remove(const Key: string);

  {  Clear   InnerKeyList     
    Count.    OwnValues  True,   Free
         . ,  
     ,     Capacity  . }

    procedure Clear;
  end;

  
{  TLinkedList -       TObject.
      ,   
      . }

  TLinkedList = class;

{  TLinkedListNode   , .. , 
    TLinkedList.  : Value -   ,
      ; OwnerList - c
     , ..  TLinkedList,  
    ; NextNode -     
   ( nil    ); PrevNode - 
       (nil    ). }

  PLinkedListNode = ^TLinkedListNode;
  TLinkedListNode = record
    Value: TObject;
    OwnerList: TLinkedList;
    NextNode: PLinkedListNode;
    PrevNode: PLinkedListNode;
  end;

  TLinkedList = class(TObject)
  private
    FHeadNode: PLinkedListNode;
    FTailNode: PLinkedListNode;
    FOwnValues: Boolean;
  public

  {  Destroy  ,    
    TLinkedListNode        
    TLinkedList.  ,   OwnValues  True,  
          Free. }

    destructor Destroy; override;

  {  }

  {  HeadNode     , .. ,
     .   ,     nil. }

    property HeadNode: PLinkedListNode read FHeadNode;

  {  TailNode     , .. ,
     .   ,     nil. }

    property TailNode: PLinkedListNode read FTailNode;

  {   OwnValues  True,      
    ,    ,    
      Free,   ,  
    .      False   Free  
        . }

    property OwnValues: Boolean read FOwnValues write FOwnValues;

  {  }

  {  AddHead      Value   
             . }

    function AddHead(Value: TObject): PLinkedListNode;

  {  AddTail      Value   
             . }

    function AddTail(Value: TObject): PLinkedListNode;

  {  InsertBefore      Value  
    Node         
     .   Node    , 
    . }

    function InsertBefore(Node: PLinkedListNode; Value: TObject): PLinkedListNode;

  {  InsertAfter      Value  
    Node         
     .   Node    , 
    . }

    function InsertAfter(Node: PLinkedListNode; Value: TObject): PLinkedListNode;

  {  IsEmpty   True,    .
           ,   False. }

    function IsEmpty: Boolean;

  {  PopHeadValue     ,  
         .    
    Free     . }

    function PopHeadValue: TObject;

  {  PopTailValue     ,  
         .    Free  
      . }

    function PopTailValue: TObject;

  {  RemoveHead      .  
    OwnValues  True,       Free. }

    procedure RemoveHead;

  {  RemoveTail      . 
     OwnValues  True,     
     Free. }

    procedure RemoveTail;

  {  Remove      Node.  
    OwnValues  True,       Free. }

    procedure Remove(Node: PLinkedListNode);

  {  Clear      .  
    OwnValues  True,       Free. }

    procedure Clear;
  end;

  
{  TArrayList -    Pointer. }

{  ,      
  .     Item1  Item2   
   ,     ,   ,
      ,  ,   . }

  TCompareFunction = function (Item1, Item2: Pointer): Integer;

{  ,      .
     ,    
   Value,   Value1  Value2  
  ,     Item.  Value ,
   ,    Item,   
   .  Value ,  , 
    Item,     . 
  Value  ,    Item,  
   .   TMatch2Function  
     ,  ,   ,  . }

  TMatchFunction = function (Value, Item: Pointer): Integer;
  TMatch2Function = function (Value1, Value2, Item: Pointer): Integer;

  TGroupEnumerator = class;

  TArrayList = class(TObject)
  private
    FCapacity: Integer;
    FItems: PPointerItemList;
    FCount: Integer;
    procedure SetCapacity(NewCapacity: Integer);
  public

  {   Create    TArrayList. 
         .   
         . }

    constructor Create; overload;

  {   Create    TArrayList  
      InitialCapacity    . }

    constructor Create(InitialCapacity: Integer); overload;

  {  Destroy  ,   
    TArrayList   ,   ItemList. 
          . }

    destructor Destroy; override;

  {  }

  {  Capacity     ,  
        .    
         Capacity, 
       Capacity   , 
     G_EnlargeCapacity(Capacity). }

    property Capacity: Integer read FCapacity write SetCapacity;

  {  Count      
     .       
     . }

    property Count: Integer read FCount write FCount;

  {  ItemList     , 
       .      
    Capacity.    ItemList    
      . }

    property ItemList: PPointerItemList read FItems;

  {  }

  {   Load    ,  
    ArrayList. }

    procedure Load(ArrayList: TArrayList); overload;

  {   Load       Pointer,
      Items,   Count . }

    procedure Load(Items: Pointer; Count: Integer); overload;

  {   EnsureCapacity ,    ,
      ItemList,     Capacity
    , ..  Capacity    TArrayList
        ,   Capacity. }

    procedure EnsureCapacity(Capacity: Integer);

  {  Add  ,   P,   . }

    procedure Add(P: Pointer);

  {  Insert  ,   P,   Index
     (  ). }

    procedure Insert(Index: Integer; P: Pointer);

  {  Peek       
    (Count - 1). }

    function Peek: Pointer;

  {  Pop  ,     
         .    
      .       Add. }

    function Pop: Pointer;

  {  IndexOf    ,  ()
      P.      ,   -1. }

    function IndexOf(P: Pointer): Integer;

  {  RemoveAt    ,    Index
    (  ). }

    procedure RemoveAt(Index: Integer);

  {  RemoveAndFreeItemAt    ,   
    Index (  ).       TObject
         Free. }

    procedure RemoveAndFreeItemAt(Index: Integer);

  {  UnorderedRemoveAt    ,   
    Index (  ).     
    ,    Count   1.  
     ,  RemoveAt,     
      . }

    procedure UnorderedRemoveAt(Index: Integer);

  {  UnorderedRemoveAt    ,   
    Index (  ).       TObject
         Free.    
     ,    Count   1. 
     ,  RemoveAndFreeItemAt,    
       . }

    procedure UnorderedRemoveAndFreeItemAt(Index: Integer);

  {  Clear       Count. ,
       ,      
     Capacity  . }

    procedure Clear;

  {  ClearAndFreeItems      
    Count.         
    TObject      Free. ,   
    ,       Capacity  . }

    procedure ClearAndFreeItems;

  {  Sort      .  
        CompareFunction. }

    procedure Sort(CompareFunction: TCompareFunction);

  {  EnumerateGroups     .
          ,  
     CompareFunction.    , 
         CompareFunction. EnumerateGroups 
      TGroupEnumerator,    . }

    function EnumerateGroups(CompareFunction: TCompareFunction): TGroupEnumerator;

  {   Search     Value 
      (Value1, Value2)  ,  
        , ,   Sort.
            .  
          MatchFunction.
        Search    . 
      ,   nil. }

    function Search(Value: Pointer; MatchFunction: TMatchFunction): Pointer; overload;
    function Search(Value1, Value2: Pointer; MatchFunction: TMatch2Function): Pointer; overload;

  {   Search     Value 
      (Value1, Value2)  ,  
        , ,   Sort.
            .  
          MatchFunction.
        Search   True   
       Index.    ,  
    False     ,     
         ,   Index. }

    function Search(Value: Pointer; MatchFunction: TMatchFunction; out Index: Integer): Boolean; overload;
    function Search(Value1, Value2: Pointer; MatchFunction: TMatch2Function; out Index: Integer): Boolean; overload;

  {  TrimToSize    ,  
    ItemList,  ,      
      . }

    procedure TrimToSize;

  {  Equals  True,     
      ArrayList.      
       ,   False. }

    function Equals(ArrayList: TArrayList): Boolean;

  {  Clone    TArrayList,  
      . }

    function Clone: TArrayList;
  end;


{  TArrayReadOnlyList      Pointer,
   TArrayList,    / . }

  TArrayReadOnlyList = class(TObject)
  private
    FItems: PPointerItemList;
    FCount: Integer;
  public

  {  Create    TArrayReadOnlyList  
     ,   ItemList,  Count . }

    constructor Create(ItemList: PPointerItemList; Count: Integer);

  {  }

  {  Count     . }

    property Count: Integer read FCount;

  {  ItemList     , 
     .    ItemList   
       . }

    property ItemList: PPointerItemList read FItems;

  {  }

  {  IndexOf    ,  
     P.      ,   -1. }

    function IndexOf(P: Pointer): Integer;

  {  Sort      .  
        CompareFunction. }

    procedure Sort(CompareFunction: TCompareFunction);

  {  EnumerateGroups     .
          ,  
     CompareFunction.    , 
         CompareFunction. EnumerateGroups 
      TGroupEnumerator,    . }

    function EnumerateGroups(CompareFunction: TCompareFunction): TGroupEnumerator;

  {   Search     Value 
      (Value1, Value2)  ,  
        , ,   Sort.
            .  
          MatchFunction.
        Search    . 
      ,   nil. }

    function Search(Value: Pointer; MatchFunction: TMatchFunction): Pointer; overload;
    function Search(Value1, Value2: Pointer; MatchFunction: TMatch2Function): Pointer; overload;

  {   Search     Value 
      (Value1, Value2)  ,  
        , ,   Sort.
            .  
          MatchFunction.
        Search   True   
       Index.    ,  
    False     ,     
         ,   Index. }

    function Search(Value: Pointer; MatchFunction: TMatchFunction; out Index: Integer): Boolean; overload;
    function Search(Value1, Value2: Pointer; MatchFunction: TMatch2Function; out Index: Integer): Boolean; overload;

  {  Equals  True,     
      ArrayList.      
       ,   False. }

    function Equals(ArrayList: TArrayList): Boolean; overload;
    function Equals(ArrayList: TArrayReadOnlyList): Boolean; overload;

  {  Clone    TArrayList,  
      . }

    function Clone: TArrayList;
  end;

  
{  TGroupEnumerator -   . }

{       ,    
     TArrayReadOnlyList. }

  PGroupList = ^TGroupList;
  TGroupList = array[0..DWordListUpperLimit] of TArrayReadOnlyList;

  TGroupEnumerator = class(TObject)
  private
    FGroups: PGroupList;
    FGroupCount: Integer;
  public

  {  Create    TGroupEnumerator, 
       .    
     ,   ItemList,   Count
    .        
     CompareFunction.     , 
        . }

    constructor Create(ItemList: PPointerItemList; Count: Integer;
      CompareFunction: TCompareFunction);

  {  Destroy  ,   
    TGroupEnumerator    ,   .
         (  TArrayReadOnlyList)
      Free. }

    destructor Destroy; override;

  {  GroupCount     . }

    property GroupCount: Integer read FGroupCount;

  {  GroupList     , 
     .    GroupList   
      . }

    property GroupList: PGroupList read FGroups;
  end;

  
{   }

{ G_SortAscending     ,
    ItemList,   Count . 
       CompareFunction. 
         CompareFunction. }

procedure G_SortAscending(ItemList: PPointerItemList; Count: Integer;
  CompareFunction: TCompareFunction);

{ G_SortDescending     ,
    ItemList,   Count . 
       CompareFunction. 
         CompareFunction. }

procedure G_SortDescending(ItemList: PPointerItemList; Count: Integer;
  CompareFunction: TCompareFunction);

implementation

uses Windows, AcedBinary, AcedCommon, AcedStrings;

{ TBitList }

constructor TBitList.Create(Count: Integer);
begin
  FCapacity := (Count + 31) shr 5;
  if FCapacity > 0 then
    FItems := G_AllocMem(FCapacity * SizeOf(LongWord));
  FCount := Count;
end;

destructor TBitList.Destroy;
begin
  if FCapacity > 0 then
    FreeMem(FItems);
end;

procedure TBitList.SetCount(NewCount: Integer);
var
  NewItemList: PLongWordItemList;
  NewCapacity, N: Integer;
begin
  NewCapacity := (NewCount + 31) shr 5;
  if NewCapacity <> FCapacity then
  begin
    if NewCapacity > 0 then
    begin
      GetMem(NewItemList, NewCapacity * SizeOf(LongWord));
      N := 0;
      if FCount > 0 then
      begin
        if FCount > NewCount then
          FCount := NewCount;
        N := (FCount + 31) shr 5;
        G_CopyLongs(FItems, NewItemList, N);
      end;
      G_FillLongs(0, @NewItemList^[N], NewCapacity - N);
    end else
      NewItemList := nil;
    if FCapacity > 0 then
      FreeMem(FItems);
    FCapacity := NewCapacity;
    FItems := NewItemList;
  end;
  FCount := NewCount;
end;

function TBitList.GetItem(Index: Integer): Boolean;
asm
        MOV     EAX,[EAX].TBitList.FItems
        BT      [EAX],EDX
        SETC    AL
end;

procedure TBitList.SetItem(Index: Integer; V: Boolean);
asm
        MOV     EAX,[EAX].TBitList.FItems
        TEST    CL,CL
        JE      @@fl
        BTS     [EAX],EDX
        RET
@@fl:   BTR     [EAX],EDX
end;

procedure TBitList.Load(Reader: TBinaryReader);
var
  C: Integer;
begin
  C := Reader.ReadWord;
  if C = $FFFF then
    C := Reader.ReadInteger;
  if (C + 31) shr 5 <> FCapacity then
  begin
    FCount := 0;
    SetCount(C);
  end else
    FCount := C;
  if C > 0 then
    Reader.Read(FItems, (C + 7) shr 3);
end;

procedure TBitList.Load(BitList: TBitList);
var
  C: Integer;
begin
  C := BitList.FCapacity;
  if C <> FCapacity then
  begin
    FCount := 0;
    SetCount(BitList.FCount);
  end else
    FCount := BitList.FCount;
  if C > 0 then
    G_CopyLongs(BitList.FItems, FItems, C);
end;

procedure TBitList.Load(Items: Pointer; Count: Integer);
var
  C: Integer;
begin
  C := (Count + 31) shr 5;
  if C <> FCapacity then
  begin
    FCount := 0;
    SetCount(Count);
  end else
    FCount := Count;
  if C > 0 then
    G_CopyLongs(Items, FItems, C);
end;

procedure TBitList.Load(Registry: TRegistry; const ValueName: string;
  EncryptionKey: PSHA256Digest);
var
  L: Integer;
  P: Pointer;
  BR: TBinaryReader;
begin
  L := Registry.GetDataSize(ValueName);
  if L > 0 then
  begin
    GetMem(P, L);
    Registry.ReadBinaryData(ValueName, P^, L);
    BR := TBinaryReader.Create;
    BR.LoadFromArray(P, L, True, EncryptionKey);
    Load(BR);
    BR.Free;
  end else
    SetCount(0);
end;

procedure TBitList.Save(Writer: TBinaryWriter);
var
  C: Integer;
begin
  C := FCount;
  if C < $FFFF then
    Writer.WriteWord(Word(C))
  else
  begin
    Writer.WriteWord($FFFF);
    Writer.WriteInteger(C);
  end;
  if C > 0 then
    Writer.Write(FItems, (C + 7) shr 3);
end;

procedure TBitList.Save(Registry: TRegistry; const ValueName: string;
  EncryptionKey: PSHA256Digest; CompressionMode: TCompressionMode);
var
  BW: TBinaryWriter;
  P: Pointer;
  L: Integer;
begin
  BW := TBinaryWriter.Create((FCount + 7) shr 3 + 10);
  Save(BW);
  L := BW.SaveToArray(P, EncryptionKey, CompressionMode);
  BW.Free;
  Registry.WriteBinaryData(ValueName, P^, L);
  FreeMem(P);
end;

procedure TBitList.FromBase64(const S: string; EncryptionKey: PSHA256Digest);
var
  L: Integer;
  P: Pointer;
  BR: TBinaryReader;
begin
  L := G_Base64Decode(S, nil);
  GetMem(P, L);
  G_Base64Decode(S, P);
  BR := TBinaryReader.Create;
  BR.LoadFromArray(P, L, True, EncryptionKey);
  Load(BR);
  BR.Free;
end;

function TBitList.ToBase64(EncryptionKey: PSHA256Digest;
  CompressionMode: TCompressionMode): string;
var
  BW: TBinaryWriter;
  P: Pointer;
  L: Integer;
begin
  BW := TBinaryWriter.Create((FCount + 7) shr 3 + 10);
  Save(BW);
  L := BW.SaveToArray(P, EncryptionKey, CompressionMode);
  BW.Free;
  Result := G_Base64Encode(P, L);
  FreeMem(P);
end;

function TBitList.Invert(Index: Integer): Boolean;
asm
        MOV     EAX,[EAX].TBitList.FItems
        BTC     [EAX],EDX
        SETC    AL
end;

function TBitList.IndexOf(V: Boolean; StartIndex: Integer): Integer;
begin
  if V then
    Result := G_SetBitScanForward(FItems, StartIndex, FCount - 1)
  else
    Result := G_FreeBitScanForward(FItems, StartIndex, FCount - 1)
end;

function TBitList.LastIndexOf(V: Boolean; StartIndex: Integer): Integer;
var
  LastBit: Integer;
begin
  if StartIndex < FCount then
    LastBit := StartIndex
  else
    LastBit := FCount - 1;
  if V then
    Result := G_SetBitScanReverse(FItems, 0, LastBit)
  else
    Result := G_FreeBitScanReverse(FItems, 0, LastBit);
end;

function TBitList.CountOf(V: Boolean): Integer;
begin
  if V then
    Result := G_CountOfSetBits(FItems, 0, FCount - 1)
  else
    Result := G_CountOfFreeBits(FItems, 0, FCount - 1);
end;

function TBitList.SetAll(V: Boolean): TBitList;
begin
  if not V then
    G_ZeroMem(FItems, (FCount + 31) and $7FFFFFE0)
  else
    G_FillLongs($FFFFFFFF, FItems, (FCount + 31) shr 5);
  Result := Self;
end;

function TBitList.NotSet: TBitList;
begin
  G_NotLongs(FItems, (FCount + 31) shr 5);
  Result := Self;
end;

function TBitList.AndSet(BitList: TBitList): TBitList;
begin
  if BitList.FCount = FCount then
    G_AndLongs(FItems, BitList.FItems, (FCount + 31) shr 5)
  else
    RaiseError(SErrCountDiffersInTBitLists);
  Result := Self;
end;

function TBitList.OrSet(BitList: TBitList): TBitList;
begin
  if BitList.FCount = FCount then
    G_OrLongs(FItems, BitList.FItems, (FCount + 31) shr 5)
  else
    RaiseError(SErrCountDiffersInTBitLists);
  Result := Self;
end;

function TBitList.XorSet(BitList: TBitList): TBitList;
begin
  if BitList.FCount = FCount then
    G_XorLongs(FItems, BitList.FItems, (FCount + 31) shr 5)
  else
    RaiseError(SErrCountDiffersInTBitLists);
  Result := Self;
end;

function TBitList.AndNotSet(BitList: TBitList): TBitList;
begin
  if BitList.FCount = FCount then
    G_AndNotLongs(FItems, BitList.FItems, (FCount + 31) shr 5)
  else
    RaiseError(SErrCountDiffersInTBitLists);
  Result := Self;
end;

function TBitList.Equals(BitList: TBitList): Boolean;
var
  Count: Integer;
begin
  Result := True;
  if BitList.FCount <> FCount then
    Result := False
  else if (BitList <> Self) and (FCount > 0) then
  begin
    Count := ((FCount + 31) shr 5) - 1;
    Result := G_SameLongs(BitList.FItems, FItems, Count) and
      (((BitList.FItems^[Count] xor FItems^[Count]) and ((1 shl (FCount - (Count shl 5))) - 1)) = 0);
  end;
end;

function TBitList.Clone: TBitList;
begin
  Result := TBitList.Create(FCount);
  G_CopyLongs(FItems, Result.FItems, Result.FCapacity);
end;

{ TIntegerList }

constructor TIntegerList.Create;
begin
  Create(0);
end;

constructor TIntegerList.Create(InitialCapacity: Integer);
begin
  FCapacity := InitialCapacity;
  if InitialCapacity > 0 then
    GetMem(FItems, InitialCapacity * SizeOf(Integer));
  FMaintainSorted := True;
  FSorted := True;
end;

destructor TIntegerList.Destroy;
begin
  if FCapacity > 0 then
    FreeMem(FItems);
end;

procedure TIntegerList.SetCapacity(NewCapacity: Integer);
var
  NewItems: PIntegerItemList;
begin
  if (NewCapacity <> FCapacity) and (NewCapacity >= FCount) then
  begin
    if NewCapacity > 0 then
    begin
      GetMem(NewItems, NewCapacity * SizeOf(Integer));
      if FCount > 0 then
        G_CopyLongs(FItems, NewItems, FCount);
    end else
      NewItems := nil;
    if FCapacity > 0 then
      FreeMem(FItems);
    FCapacity := NewCapacity;
    FItems := NewItems;
  end;
end;

procedure TIntegerList.SetMaintainSorted(Value: Boolean);
begin
  FMaintainSorted := Value;
  if Value then
    EnsureSorted;
end;

procedure TIntegerList.Load(Reader: TBinaryReader);
var
  C: Integer;
begin
  C := Reader.ReadSmallInt;
  if C = -32768 then
    C := Reader.ReadInteger;
  if C >= 0 then
    FSorted := True
  else
  begin
    FSorted := False;
    C := -C;
  end;
  if C > FCapacity then
  begin
    FCount := 0;
    SetCapacity(G_NormalizeCapacity(C));
  end;
  if C > 0 then
    Reader.Read(FItems, C * SizeOf(Integer));
  FCount := C;
  if FMaintainSorted then
    EnsureSorted;
end;

procedure TIntegerList.Load(IntegerList: TIntegerList);
var
  C: Integer;
begin
  C := IntegerList.FCount;
  if C > FCapacity then
  begin
    FCount := 0;
    SetCapacity(G_NormalizeCapacity(C));
  end;
  if C > 0 then
    G_CopyLongs(IntegerList.FItems, FItems, C);
  FCount := C;
  FSorted := IntegerList.FSorted;
  if FMaintainSorted then
    EnsureSorted;
end;

procedure TIntegerList.Load(Items: Pointer; Count: Integer; Sorted: Boolean);
begin
  if Count > FCapacity then
  begin
    FCount := 0;
    SetCapacity(G_NormalizeCapacity(Count));
  end;
  if Count > 0 then
    G_CopyLongs(Items, FItems, Count);
  FSorted := Sorted;
  FCount := Count;
  if FMaintainSorted then
    EnsureSorted;
end;

procedure TIntegerList.Load(Registry: TRegistry; const ValueName: string;
  EncryptionKey: PSHA256Digest);
var
  L: Integer;
  P: Pointer;
  BR: TBinaryReader;
begin
  L := Registry.GetDataSize(ValueName);
  if L > 0 then
  begin
    GetMem(P, L);
    Registry.ReadBinaryData(ValueName, P^, L);
    BR := TBinaryReader.Create;
    BR.LoadFromArray(P, L, True, EncryptionKey);
    Load(BR);
    BR.Free;
  end else
    Clear;
end;

procedure TIntegerList.Save(Writer: TBinaryWriter);
var
  C: Integer;
begin
  C := FCount;
  if C < 32768 then
  begin
    if FSorted then
      Writer.WriteSmallInt(SmallInt(C))
    else
      Writer.WriteSmallInt(-SmallInt(C))
  end else
  begin
    Writer.WriteSmallInt(-32768);
    if FSorted then
      Writer.WriteInteger(C)
    else
      Writer.WriteInteger(-C)
  end;
  if C > 0 then
    Writer.Write(FItems, C * SizeOf(Integer));
end;

procedure TIntegerList.Save(Registry: TRegistry; const ValueName: string;
  EncryptionKey: PSHA256Digest; CompressionMode: TCompressionMode);
var
  BW: TBinaryWriter;
  P: Pointer;
  L: Integer;
begin
  BW := TBinaryWriter.Create(FCount * SizeOf(Integer) + 10);
  Save(BW);
  L := BW.SaveToArray(P, EncryptionKey, CompressionMode);
  BW.Free;
  Registry.WriteBinaryData(ValueName, P^, L);
  FreeMem(P);
end;

procedure TIntegerList.FromBase64(const S: string; EncryptionKey: PSHA256Digest);
var
  L: Integer;
  P: Pointer;
  BR: TBinaryReader;
begin
  L := G_Base64Decode(S, nil);
  GetMem(P, L);
  G_Base64Decode(S, P);
  BR := TBinaryReader.Create;
  BR.LoadFromArray(P, L, True, EncryptionKey);
  Load(BR);
  BR.Free;
end;

function TIntegerList.ToBase64(EncryptionKey: PSHA256Digest;
  CompressionMode: TCompressionMode): string;
var
  BW: TBinaryWriter;
  P: Pointer;
  L: Integer;
begin
  BW := TBinaryWriter.Create(FCount * SizeOf(Integer) + 10);
  Save(BW);
  L := BW.SaveToArray(P, EncryptionKey, CompressionMode);
  BW.Free;
  Result := G_Base64Encode(P, L);
  FreeMem(P);
end;

procedure TIntegerList.EnsureCapacity(Capacity: Integer);
begin
  if FCapacity < Capacity then
    SetCapacity(G_NormalizeCapacity(Capacity));
end;

procedure TIntegerList.Add(N: Integer);
var
  L: Integer;
begin
  if FCount >= FCapacity then
    SetCapacity(G_EnlargeCapacity(FCapacity));
  if FMaintainSorted then
  begin
    L := G_SearchFirstGE_Integer(N, FItems, FCount);
    if L < FCount then
      G_MoveLongs(@FItems^[L], @FItems^[L + 1], FCount - L);
    FItems^[L] := N;
  end else
  begin
    FItems^[FCount] := N;
    FSorted := False;
  end;
  Inc(FCount);
end;

procedure TIntegerList.AddIfNotExists(N: Integer);
var
  L: Integer;
begin
  if FSorted then
  begin
    L := G_SearchFirstGE_Integer(N, FItems, FCount);
    if (L < FCount) and (FItems^[L] = N) then
      Exit;
  end else
  begin
    L := G_Scan_Integer(N, FItems, FCount);
    if L >= 0 then
      Exit;
  end;
  if FCount >= FCapacity then
    SetCapacity(G_EnlargeCapacity(FCapacity));
  if FMaintainSorted then
  begin
    if L < FCount then
      G_MoveLongs(@FItems^[L], @FItems^[L + 1], FCount - L);
    FItems^[L] := N;
  end else
  begin
    FItems^[FCount] := N;
    FSorted := False;
  end;
  Inc(FCount);
end;

procedure TIntegerList.Insert(Index, N: Integer);
begin
  if FMaintainSorted then
    RaiseError(SErrInsertionNotAllowed);
  if FCount >= FCapacity then
    SetCapacity(G_EnlargeCapacity(FCapacity));
  if Index < FCount then
    G_MoveLongs(@FItems^[Index], @FItems^[Index + 1], FCount - Index);
  FItems^[Index] := N;
  Inc(FCount);
  FSorted := False;
end;

function TIntegerList.Peek: Integer;
begin
  Result := FItems^[FCount - 1];
end;

function TIntegerList.Pop: Integer;
begin
  Dec(FCount);
  Result := FItems^[FCount];
end;

function TIntegerList.IndexOf(N: Integer): Integer;
begin
  if FSorted then
    Result := G_SearchUnique_Integer(N, FItems, FCount)
  else
    Result := G_Scan_Integer(N, FItems, FCount);
end;

procedure TIntegerList.Remove(N: Integer);
var
  L: Integer;
begin
  if FSorted then
    L := G_SearchUnique_Integer(N, FItems, FCount)
  else
    L := G_Scan_Integer(N, FItems, FCount);
  if L >= 0 then
    RemoveAt(L);
end;

procedure TIntegerList.RemoveAt(Index: Integer);
begin
  Dec(FCount);
  if Index < FCount then
    G_MoveLongs(@FItems^[Index + 1], @FItems^[Index], FCount - Index)
end;

procedure TIntegerList.UnorderedRemove(N: Integer);
var
  L: Integer;
begin
  if FSorted then
    L := G_SearchUnique_Integer(N, FItems, FCount)
  else
    L := G_Scan_Integer(N, FItems, FCount);
  if L >= 0 then
    UnorderedRemoveAt(L);
end;

procedure TIntegerList.UnorderedRemoveAt(Index: Integer);
begin
  if FMaintainSorted then
    RaiseError(SErrUnorderedRemovalNotAllowed);
  Dec(FCount);
  if Index < FCount then
  begin
    FItems^[Index] := FItems^[FCount];
    FSorted := False;
  end;
end;

procedure TIntegerList.Clear;
begin
  FCount := 0;
end;

procedure TIntegerList.EnsureSorted;
begin
  if not FSorted then
  begin
    if FCount > 1 then
      G_Sort_Integer(FItems, FCount);
    FSorted := True;
  end;
end;

procedure TIntegerList.TrimToSize;
begin
  if FCount < FCapacity then
    SetCapacity(FCount);
end;

function TIntegerList.Equals(IntegerList: TIntegerList): Boolean;
begin
  if IntegerList = Self then
    Result := True
  else if (IntegerList <> nil) and (IntegerList.FCount = FCount) then
    Result := G_SameLongs(FItems, IntegerList.FItems, FCount)
  else
    Result := False;
end;

function TIntegerList.Clone: TIntegerList;
begin
  Result := TIntegerList.Create(FCount);
  if FCount > 0 then
    G_CopyLongs(FItems, Result.FItems, FCount);
  Result.FCount := FCount;
  Result.FMaintainSorted := FMaintainSorted;
  Result.FSorted := FSorted;
end;

{ TWordList }

constructor TWordList.Create;
begin
  Create(0);
end;

constructor TWordList.Create(InitialCapacity: Integer);
begin
  FCapacity := InitialCapacity;
  if InitialCapacity > 0 then
    GetMem(FItems, InitialCapacity * SizeOf(Word));
  FMaintainSorted := True;
  FSorted := True;
end;

destructor TWordList.Destroy;
begin
  if FCapacity > 0 then
    FreeMem(FItems);
end;

procedure TWordList.SetCapacity(NewCapacity: Integer);
var
  NewItems: PWordItemList;
begin
  if (NewCapacity <> FCapacity) and (NewCapacity >= FCount) then
  begin
    if NewCapacity > 0 then
    begin
      GetMem(NewItems, NewCapacity * SizeOf(Word));
      if FCount > 0 then
        G_CopyMem(FItems, NewItems, FCount * SizeOf(Word));
    end else
      NewItems := nil;
    if FCapacity > 0 then
      FreeMem(FItems);
    FCapacity := NewCapacity;
    FItems := NewItems;
  end;
end;

procedure TWordList.SetMaintainSorted(Value: Boolean);
begin
  FMaintainSorted := Value;
  if Value then
    EnsureSorted;
end;

procedure TWordList.Load(Reader: TBinaryReader);
var
  C: Integer;
begin
  C := Reader.ReadSmallInt;
  if C = -32768 then
    C := Reader.ReadInteger;
  if C >= 0 then
    FSorted := True
  else
  begin
    FSorted := False;
    C := -C;
  end;
  if C > FCapacity then
  begin
    FCount := 0;
    SetCapacity(G_NormalizeCapacity(C));
  end;
  if C > 0 then
    Reader.Read(FItems, C * SizeOf(Word));
  FCount := C;
  if FMaintainSorted then
    EnsureSorted;
end;

procedure TWordList.Load(WordList: TWordList);
var
  C: Integer;
begin
  C := WordList.FCount;
  if C > FCapacity then
  begin
    FCount := 0;
    SetCapacity(G_NormalizeCapacity(C));
  end;
  if C > 0 then
    G_CopyMem(WordList.FItems, FItems, C * SizeOf(Word));
  FCount := C;
  FSorted := WordList.FSorted;
  if FMaintainSorted then
    EnsureSorted;
end;

procedure TWordList.Load(Items: Pointer; Count: Integer; Sorted: Boolean);
begin
  if Count > FCapacity then
  begin
    FCount := 0;
    SetCapacity(G_NormalizeCapacity(Count));
  end;
  if Count > 0 then
    G_CopyMem(Items, FItems, Count * SizeOf(Word));
  FSorted := Sorted;
  FCount := Count;
  if FMaintainSorted then
    EnsureSorted;
end;

procedure TWordList.Load(Registry: TRegistry; const ValueName: string;
  EncryptionKey: PSHA256Digest);
var
  L: Integer;
  P: Pointer;
  BR: TBinaryReader;
begin
  L := Registry.GetDataSize(ValueName);
  if L > 0 then
  begin
    GetMem(P, L);
    Registry.ReadBinaryData(ValueName, P^, L);
    BR := TBinaryReader.Create;
    BR.LoadFromArray(P, L, True, EncryptionKey);
    Load(BR);
    BR.Free;
  end else
    Clear;
end;

procedure TWordList.Save(Writer: TBinaryWriter);
var
  C: Integer;
begin
  C := FCount;
  if C < 32768 then
  begin
    if FSorted then
      Writer.WriteSmallInt(SmallInt(C))
    else
      Writer.WriteSmallInt(-SmallInt(C))
  end else
  begin
    Writer.WriteSmallInt(-32768);
    if FSorted then
      Writer.WriteInteger(C)
    else
      Writer.WriteInteger(-C)
  end;
  if C > 0 then
    Writer.Write(FItems, C * SizeOf(Word));
end;

procedure TWordList.Save(Registry: TRegistry; const ValueName: string;
  EncryptionKey: PSHA256Digest; CompressionMode: TCompressionMode);
var
  BW: TBinaryWriter;
  P: Pointer;
  L: Integer;
begin
  BW := TBinaryWriter.Create(FCount * SizeOf(Word) + 10);
  Save(BW);
  L := BW.SaveToArray(P, EncryptionKey, CompressionMode);
  BW.Free;
  Registry.WriteBinaryData(ValueName, P^, L);
  FreeMem(P);
end;

procedure TWordList.FromBase64(const S: string; EncryptionKey: PSHA256Digest);
var
  L: Integer;
  P: Pointer;
  BR: TBinaryReader;
begin
  L := G_Base64Decode(S, nil);
  GetMem(P, L);
  G_Base64Decode(S, P);
  BR := TBinaryReader.Create;
  BR.LoadFromArray(P, L, True, EncryptionKey);
  Load(BR);
  BR.Free;
end;

function TWordList.ToBase64(EncryptionKey: PSHA256Digest;
  CompressionMode: TCompressionMode): string;
var
  BW: TBinaryWriter;
  P: Pointer;
  L: Integer;
begin
  BW := TBinaryWriter.Create(FCount * SizeOf(Word) + 10);
  Save(BW);
  L := BW.SaveToArray(P, EncryptionKey, CompressionMode);
  BW.Free;
  Result := G_Base64Encode(P, L);
  FreeMem(P);
end;

procedure TWordList.EnsureCapacity(Capacity: Integer);
begin
  if FCapacity < Capacity then
    SetCapacity(G_NormalizeCapacity(Capacity));
end;

procedure TWordList.Add(N: Word);
var
  L: Integer;
begin
  if FCount >= FCapacity then
    SetCapacity(G_EnlargeCapacity(FCapacity));
  if FMaintainSorted then
  begin
    L := G_SearchFirstGE_Word(N, FItems, FCount);
    if L < FCount then
      G_MoveWords(@FItems^[L], @FItems^[L + 1], FCount - L);
    FItems^[L] := N;
  end else
  begin
    FItems^[FCount] := N;
    FSorted := False;
  end;
  Inc(FCount);
end;

procedure TWordList.AddIfNotExists(N: Word);
var
  L: Integer;
begin
  if FSorted then
  begin
    L := G_SearchFirstGE_Word(N, FItems, FCount);
    if (L < FCount) and (FItems^[L] = N) then
      Exit;
  end else
  begin
    L := G_Scan_Word(N, FItems, FCount);
    if L >= 0 then
      Exit;
  end;
  if FCount >= FCapacity then
    SetCapacity(G_EnlargeCapacity(FCapacity));
  if FMaintainSorted then
  begin
    if L < FCount then
      G_MoveWords(@FItems^[L], @FItems^[L + 1], FCount - L);
    FItems^[L] := N;
  end else
  begin
    FItems^[FCount] := N;
    FSorted := False;
  end;
  Inc(FCount);
end;

procedure TWordList.Insert(Index: Integer; N: Word);
begin
  if FMaintainSorted then
    RaiseError(SErrInsertionNotAllowed);
  if FCount >= FCapacity then
    SetCapacity(G_EnlargeCapacity(FCapacity));
  if Index < FCount then
    G_MoveWords(@FItems^[Index], @FItems^[Index + 1], FCount - Index);
  FItems^[Index] := N;
  Inc(FCount);
  FSorted := False;
end;

function TWordList.Peek: Word;
begin
  Result := FItems^[FCount - 1];
end;

function TWordList.Pop: Word;
begin
  Dec(FCount);
  Result := FItems^[FCount];
end;

function TWordList.IndexOf(N: Word): Integer;
begin
  if FSorted then
    Result := G_SearchUnique_Word(N, FItems, FCount)
  else
    Result := G_Scan_Word(N, FItems, FCount);
end;

procedure TWordList.Remove(N: Word);
var
  L: Integer;
begin
  if FSorted then
    L := G_SearchUnique_Word(N, FItems, FCount)
  else
    L := G_Scan_Word(N, FItems, FCount);
  if L >= 0 then
    RemoveAt(L);
end;

procedure TWordList.RemoveAt(Index: Integer);
begin
  Dec(FCount);
  if Index < FCount then
    G_MoveWords(@FItems^[Index + 1], @FItems^[Index], FCount - Index)
end;

procedure TWordList.UnorderedRemove(N: Word);
var
  L: Integer;
begin
  if FSorted then
    L := G_SearchUnique_Word(N, FItems, FCount)
  else
    L := G_Scan_Word(N, FItems, FCount);
  if L >= 0 then
    UnorderedRemoveAt(L);
end;

procedure TWordList.UnorderedRemoveAt(Index: Integer);
begin
  if FMaintainSorted then
    RaiseError(SErrUnorderedRemovalNotAllowed);
  Dec(FCount);
  if Index < FCount then
  begin
    FItems^[Index] := FItems^[FCount];
    FSorted := False;
  end;
end;

procedure TWordList.Clear;
begin
  FCount := 0;
end;

procedure TWordList.EnsureSorted;
begin
  if not FSorted then
  begin
    if FCount > 1 then
      G_Sort_Word(FItems, FCount);
    FSorted := True;
  end;
end;

procedure TWordList.TrimToSize;
begin
  if FCount < FCapacity then
    SetCapacity(FCount);
end;

function TWordList.Equals(WordList: TWordList): Boolean;
begin
  if WordList = Self then
    Result := True
  else if (WordList <> nil) and (WordList.FCount = FCount) then
    Result := G_SameMem(FItems, WordList.FItems, FCount * SizeOf(Word))
  else
    Result := False;
end;

function TWordList.Clone: TWordList;
begin
  Result := TWordList.Create(FCount);
  if FCount > 0 then
    G_CopyMem(FItems, Result.FItems, FCount * SizeOf(Word));
  Result.FCount := FCount;
  Result.FMaintainSorted := FMaintainSorted;
  Result.FSorted := FSorted;
end;

{ TIntegerAssociatedList }

constructor TIntegerAssociatedList.Create;
begin
  Create(0);
end;

constructor TIntegerAssociatedList.Create(InitialCapacity: Integer);
begin
  FCapacity := InitialCapacity;
  if InitialCapacity > 0 then
  begin
    GetMem(FKeys, InitialCapacity * SizeOf(Integer));
    GetMem(FValues, InitialCapacity * SizeOf(Pointer));
  end;
end;

destructor TIntegerAssociatedList.Destroy;
var
  I: Integer;
begin
  if FCapacity > 0 then
  begin
    if FOwnValues then
      for I := FCount - 1 downto 0 do
        FValues^[I].Free;
    FreeMem(FKeys);
    FreeMem(FValues);
  end;
end;

procedure TIntegerAssociatedList.SetCapacity(NewCapacity: Integer);
var
  NewKeys: PIntegerItemList;
  NewValues: PObjectItemList;
begin
  if (NewCapacity <> FCapacity) and (NewCapacity >= FCount) then
  begin
    if NewCapacity > 0 then
    begin
      GetMem(NewKeys, NewCapacity * SizeOf(Integer));
      GetMem(NewValues, NewCapacity * SizeOf(Pointer));
      if FCount > 0 then
      begin
        G_CopyLongs(FKeys, NewKeys, FCount);
        G_CopyLongs(FValues, NewValues, FCount);
      end;
    end else
    begin
      NewKeys := nil;
      NewValues := nil;
    end;
    if FCapacity > 0 then
    begin
      FreeMem(FKeys);
      FreeMem(FValues);
    end;
    FCapacity := NewCapacity;
    FKeys := NewKeys;
    FValues := NewValues;
  end;
end;

function TIntegerAssociatedList.GetItem(Key: Integer): TObject;
var
  L: Integer;
begin
  L := G_SearchUnique_Integer(Key, FKeys, FCount);
  if L >= 0 then
    Result := FValues^[L]
  else
    Result := nil;
end;

procedure TIntegerAssociatedList.SetItem(Key: Integer; Value: TObject);
var
  L: Integer;
  O: TObject;
begin
  L := G_SearchFirstGE_Integer(Key, FKeys, FCount);
  if (L < FCount) and (FKeys^[L] = Key) then
  begin
    if not FOwnValues then
      FValues^[L] := Value
    else
    begin
      O := FValues^[L];
      if O <> Value then
      begin
        FValues^[L] := Value;
        O.Free;
      end;
    end;
  end else
  begin
    if FCount >= FCapacity then
      SetCapacity(G_EnlargeCapacity(FCapacity));
    if L < FCount then
    begin
      G_MoveLongs(@FKeys^[L], @FKeys^[L + 1], FCount - L);
      G_MoveLongs(@FValues^[L], @FValues^[L + 1], FCount - L);
    end;
    FKeys^[L] := Key;
    FValues^[L] := Value;
    Inc(FCount);
  end;
end;

procedure TIntegerAssociatedList.EnsureCapacity(Capacity: Integer);
begin
  if FCapacity < Capacity then
    SetCapacity(G_NormalizeCapacity(Capacity));
end;

procedure TIntegerAssociatedList.Add(Key: Integer; Value: TObject);
var
  L: Integer;
begin
  L := G_SearchFirstGE_Integer(Key, FKeys, FCount);
  if (L < FCount) and (FKeys^[L] = Key) then
    RaiseErrorFmt(SErrKeyDuplicatesInAssociatedList, 'TIntegerAssociatedList');
  if FCount >= FCapacity then
    SetCapacity(G_EnlargeCapacity(FCapacity));
  if L < FCount then
  begin
    G_MoveLongs(@FKeys^[L], @FKeys^[L + 1], FCount - L);
    G_MoveLongs(@FValues^[L], @FValues^[L + 1], FCount - L);
  end;
  FKeys^[L] := Key;
  FValues^[L] := Value;
  Inc(FCount);
end;

function TIntegerAssociatedList.IndexOf(Key: Integer): Integer;
begin
  Result := G_SearchUnique_Integer(Key, FKeys, FCount)
end;

procedure TIntegerAssociatedList.Remove(Key: Integer);
var
  L: Integer;
begin
  L := G_SearchUnique_Integer(Key, FKeys, FCount);
  if L >= 0 then
    RemoveAt(L);
end;

procedure TIntegerAssociatedList.RemoveAt(Index: Integer);
begin
  if FOwnValues then
    FValues^[Index].Free;
  Dec(FCount);
  if Index < FCount then
  begin
    G_MoveLongs(@FKeys^[Index + 1], @FKeys^[Index], FCount - Index);
    G_MoveLongs(@FValues^[Index + 1], @FValues^[Index], FCount - Index);
  end;
end;

procedure TIntegerAssociatedList.Clear;
var
  I: Integer;
begin
  if FOwnValues then
    for I := FCount - 1 downto 0 do
      FValues^[I].Free;
  FCount := 0;
end;

procedure TIntegerAssociatedList.TrimToSize;
begin
  if FCount < FCapacity then
    SetCapacity(FCount);
end;

{ TIntegerHashTable }

const
  MissingKey  = Integer($81000000);
  EmptyKey    = Integer($81000001);

constructor TIntegerHashTable.Create;
begin
  Create(0);
end;

constructor TIntegerHashTable.Create(InitialCapacity: Integer);
var
  N: Integer;
begin
  N := G_EnlargePrimeCapacity((InitialCapacity * 7) div 5);
  FCapacity := N;
  FMaxCount := (N * 5) div 7;
  GetMem(FKeys, N shl 2);
  G_FillLongs(LongWord(EmptyKey), FKeys, N);
  GetMem(FValues, N shl 2);
  N := (N + 31) shr 5;
  GetMem(FBits, N shl 2);
  G_FillLongs(0, FBits, N);
end;

destructor TIntegerHashTable.Destroy;
var
  I: Integer;
begin
  if FOwnValues then
    for I := FCapacity - 1 downto 0 do
      if G_BitTest(FBits, I) then
        FValues^[I].Free;
  FreeMem(FKeys);
  FreeMem(FValues);
  FreeMem(FBits);
end;

procedure TIntegerHashTable.SetMaxCount(NewMaxCount: Integer);
var
  N: Integer;
begin
  if NewMaxCount >= FCount then
  begin
    N := G_EnlargePrimeCapacity((NewMaxCount * 7) div 5);
    if N <> FCapacity then
      SetCapacity(N);
  end;
end;

function TIntegerHashTable.GetIndex(Key: Integer): Integer;
var
  X, HashSize, HashStep: LongWord;
begin
  X := LongWord(Key);
  HashSize := LongWord(FCapacity);
  HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
  while True do
  begin
    X := X mod HashSize;
    if (FKeys^[X] = EmptyKey) and not G_BitTest(FBits, X) then
    begin
      Result := X;
      Exit;
    end;
    Inc(X, HashStep);
  end;
end;

procedure TIntegerHashTable.SetCapacity(NewCapacity: Integer);
var
  PrevKeys: PIntegerItemList;
  PrevValues: PObjectItemList;
  PrevBits: PLongWordItemList;
  N, I: Integer;
begin
  N := NewCapacity;
  I := FCapacity - 1;
  PrevKeys := FKeys;
  PrevValues := FValues;
  PrevBits := FBits;
  FCapacity := N;
  FMaxCount := (N * 5) div 7;
  GetMem(FKeys, N shl 2);
  G_FillLongs(LongWord(EmptyKey), FKeys, N);
  GetMem(FValues, N shl 2);
  N := (N + 31) shr 5;
  GetMem(FBits, N shl 2);
  G_FillLongs(0, FBits, N);
  while I >= 0 do
  begin
    if G_BitTest(PrevBits, I) then
    begin
      N := GetIndex(PrevKeys^[I]);
      FKeys^[N] := PrevKeys^[I];
      FValues^[N] := PrevValues^[I];
      G_BitSet(FBits, N);
    end;
    Dec(I);
  end;
  FreeMem(PrevKeys);
  FreeMem(PrevValues);
  FreeMem(PrevBits);
end;

function TIntegerHashTable.GetItem(Key: Integer): TObject;
var
  X, HashSize, HashStep: LongWord;
begin
  X := LongWord(Key);
  HashSize := LongWord(FCapacity);
  HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
  while True do
  begin
    X := X mod HashSize;
    if (FKeys^[X] = EmptyKey) and not G_BitTest(FBits, X) then
      Break;
    if (FKeys^[X] = Key) and ((Key <> MissingKey) or G_BitTest(FBits, X)) then
    begin
      Result := FValues^[X];
      Exit;
    end;
    Inc(X, HashStep);
  end;
  Result := nil;
end;

procedure TIntegerHashTable.SetItem(Key: Integer; Value: TObject);
var
  X, HashSize, HashStep: LongWord;
  L: Integer;
  O: TObject;
begin
  X := LongWord(Key);
  HashSize := LongWord(FCapacity);
  HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
  L := -1;
  while True do
  begin
    X := X mod HashSize;
    if G_BitTest(FBits, X) then
    begin
      if FKeys^[X] = Key then
      begin
        if not FOwnValues then
          FValues^[X] := Value
        else
        begin
          O := FValues^[X];
          if O <> Value then
          begin
            FValues^[X] := Value;
            O.Free;
          end;
        end;
        Exit;
      end;
    end else
    begin
      if L < 0 then
        L := X;
      if FKeys^[X] = EmptyKey then
        Break;
    end;
    Inc(X, HashStep);
  end;
  if FCount >= FMaxCount then
  begin
    SetCapacity(G_EnlargePrimeCapacity(FCapacity));
    L := GetIndex(Key);
  end;
  FKeys^[L] := Key;
  FValues^[L] := Value;
  G_BitSet(FBits, L);
  Inc(FCount);
end;

procedure TIntegerHashTable.EnsureCapacity(Capacity: Integer);
begin
  if FMaxCount < Capacity then
    SetCapacity(G_EnlargePrimeCapacity((Capacity * 7) div 5));
end;

procedure TIntegerHashTable.Add(Key: Integer; Value: TObject);
var
  X, HashSize, HashStep: LongWord;
  L: Integer;
begin
  X := LongWord(Key);
  HashSize := LongWord(FCapacity);
  HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
  L := -1;
  while True do
  begin
    X := X mod HashSize;
    if G_BitTest(FBits, X) then
    begin
      if FKeys^[X] = Key then
        RaiseErrorFmt(SErrKeyDuplicatesInAssociatedList, 'TIntegerHashTable');
    end else
    begin
      if L < 0 then
        L := X;
      if FKeys^[X] = EmptyKey then
        Break;
    end;
    Inc(X, HashStep);
  end;
  if FCount >= FMaxCount then
  begin
    SetCapacity(G_EnlargePrimeCapacity(FCapacity));
    L := GetIndex(Key);
  end;
  FKeys^[L] := Key;
  FValues^[L] := Value;
  G_BitSet(FBits, L);
  Inc(FCount);
end;

function TIntegerHashTable.Contains(Key: Integer): Boolean;
var
  X, HashSize, HashStep: LongWord;
begin
  X := LongWord(Key);
  HashSize := LongWord(FCapacity);
  HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
  while True do
  begin
    X := X mod HashSize;
    if (FKeys^[X] = EmptyKey) and not G_BitTest(FBits, X) then
      Break;
    if (FKeys^[X] = Key) and ((Key <> MissingKey) or G_BitTest(FBits, X)) then
    begin
      Result := True;
      Exit;
    end;
    Inc(X, HashStep);
  end;
  Result := False;
end;

procedure TIntegerHashTable.Remove(Key: Integer);
var
  X, HashSize, HashStep: LongWord;
begin
  X := LongWord(Key);
  HashSize := LongWord(FCapacity);
  HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
  while True do
  begin
    X := X mod HashSize;
    if (FKeys^[X] = EmptyKey) and not G_BitTest(FBits, X) then
      Break;
    if (FKeys^[X] = Key) and ((Key <> MissingKey) or G_BitTest(FBits, X)) then
    begin
      if FOwnValues then
        FValues^[X].Free;
      FKeys^[X] := MissingKey;
      G_BitReset(FBits, X);
      Dec(FCount);
      Exit;
    end;
    Inc(X, HashStep);
  end;
end;

procedure TIntegerHashTable.Clear;
var
  I, N: Integer;
begin
  N := FCapacity;
  if FOwnValues then
    for I := N - 1 downto 0 do
      if G_BitTest(FBits, I) then
        FValues^[I].Free;
  G_FillLongs(LongWord(EmptyKey), FKeys, N);
  N := (N + 31) shr 5;
  G_FillLongs(0, FBits, N);
  FCount := 0;
end;

{ TStringAssociatedList }

constructor TStringAssociatedList.Create(CaseSensitive: Boolean);
begin
  Create(CaseSensitive, 0);
end;

constructor TStringAssociatedList.Create(CaseSensitive: Boolean;
  InitialCapacity: Integer);
begin
  FCapacity := InitialCapacity;
  FCaseSensitive := CaseSensitive;
  if InitialCapacity > 0 then
  begin
    GetMem(FKeys, InitialCapacity * SizeOf(Pointer));
    GetMem(FValues, InitialCapacity * SizeOf(Pointer));
  end;
end;

destructor TStringAssociatedList.Destroy;
var
  I: Integer;
begin
  if FCapacity > 0 then
  begin
    Finalize(FKeys^[0], FCount);
    if FOwnValues then
      for I := FCount - 1 downto 0 do
        FValues^[I].Free;
    FreeMem(FKeys);
    FreeMem(FValues);
  end;
end;

procedure TStringAssociatedList.SetCapacity(NewCapacity: Integer);
var
  NewKeys: PStringItemList;
  NewValues: PObjectItemList;
begin
  if (NewCapacity <> FCapacity) and (NewCapacity >= FCount) then
  begin
    if NewCapacity > 0 then
    begin
      GetMem(NewKeys, NewCapacity * SizeOf(Pointer));
      GetMem(NewValues, NewCapacity * SizeOf(Pointer));
      if FCount > 0 then
      begin
        G_CopyLongs(FKeys, NewKeys, FCount);
        G_CopyLongs(FValues, NewValues, FCount);
      end;
    end else
    begin
      NewKeys := nil;
      NewValues := nil;
    end;
    if FCapacity > 0 then
    begin
      FreeMem(FKeys);
      FreeMem(FValues);
    end;
    FCapacity := NewCapacity;
    FKeys := NewKeys;
    FValues := NewValues;
  end;
end;

function TStringAssociatedList.GetItem(const Key: string): TObject;
var
  L: Integer;
begin
  L := IndexOf(Key);
  if L >= 0 then
    Result := FValues^[L]
  else
    Result := nil;
end;

function TStringAssociatedList.SearchFirstGE(const Key: string): Integer;
var
  L, H, I: Integer;
begin
  L := 0;
  H := FCount - 1;
  if not FCaseSensitive then
    while L <= H do
    begin
      I := (L + H) shr 1;
      if G_CompareText(FKeys^[I], Key) < 0 then
        L := I + 1
      else
        H := I - 1;
    end
  else
    while L <= H do
    begin
      I := (L + H) shr 1;
      if G_CompareStr(FKeys^[I], Key) < 0 then
        L := I + 1
      else
        H := I - 1;
    end;
  Result := L;
end;

procedure TStringAssociatedList.SetItem(const Key: string; Value: TObject);
var
  L: Integer;
  O: TObject;
begin
  L := SearchFirstGE(Key);
  if (L < FCount) and (FKeys^[L] = Key) then
  begin
    if not FOwnValues then
      FValues^[L] := Value
    else
    begin
      O := FValues^[L];
      if O <> Value then
      begin
        FValues^[L] := Value;
        O.Free;
      end;
    end;
  end else
  begin
    if FCount >= FCapacity then
      SetCapacity(G_EnlargeCapacity(FCapacity));
    if L < FCount then
    begin
      G_MoveLongs(@FKeys^[L], @FKeys^[L + 1], FCount - L);
      G_MoveLongs(@FValues^[L], @FValues^[L + 1], FCount - L);
    end;
    Pointer(FKeys^[L]) := nil;
    FKeys^[L] := Key;
    FValues^[L] := Value;
    Inc(FCount);
  end;
end;

procedure TStringAssociatedList.EnsureCapacity(Capacity: Integer);
begin
  if FCapacity < Capacity then
    SetCapacity(G_NormalizeCapacity(Capacity));
end;

procedure TStringAssociatedList.Add(const Key: string; Value: TObject);
var
  L: Integer;
begin
  L := SearchFirstGE(Key);
  if (L < FCount) and (FKeys^[L] = Key) then
    RaiseErrorFmt(SErrKeyDuplicatesInAssociatedList, 'TStringAssociatedList');
  if FCount >= FCapacity then
    SetCapacity(G_EnlargeCapacity(FCapacity));
  if L < FCount then
  begin
    G_MoveLongs(@FKeys^[L], @FKeys^[L + 1], FCount - L);
    G_MoveLongs(@FValues^[L], @FValues^[L + 1], FCount - L);
  end;
  Pointer(FKeys^[L]) := nil;
  FKeys^[L] := Key;
  FValues^[L] := Value;
  Inc(FCount);
end;

function TStringAssociatedList.IndexOf(const Key: string): Integer;
var
  L, H, I, C: Integer;
begin
  L := 0;
  H := FCount - 1;
  if not FCaseSensitive then
    while L <= H do
    begin
      I := (L + H) shr 1;
      C := G_CompareText(FKeys^[I], Key);
      if C < 0 then
        L := I + 1
      else
      begin
        if C = 0 then
        begin
          Result := I;
          Exit;
        end;
        H := I - 1;
      end;
    end
  else
    while L <= H do
    begin
      I := (L + H) shr 1;
      C := G_CompareStr(FKeys^[I], Key);
      if C < 0 then
        L := I + 1
      else
      begin
        if C = 0 then
        begin
          Result := I;
          Exit;
        end;
        H := I - 1;
      end;
    end;
  Result := -1;
end;

procedure TStringAssociatedList.Remove(const Key: string);
var
  L: Integer;
begin
  L := IndexOf(Key);
  if L >= 0 then
    RemoveAt(L);
end;

procedure TStringAssociatedList.RemoveAt(Index: Integer);
begin
  FKeys^[Index] := '';
  if FOwnValues then
    FValues^[Index].Free;
  Dec(FCount);
  if Index < FCount then
  begin
    G_MoveLongs(@FKeys^[Index + 1], @FKeys^[Index], FCount - Index);
    G_MoveLongs(@FValues^[Index + 1], @FValues^[Index], FCount - Index);
  end;
end;

procedure TStringAssociatedList.Clear;
var
  I: Integer;
begin
  Finalize(FKeys^[0], FCount);
  if FOwnValues then
    for I := FCount - 1 downto 0 do
      FValues^[I].Free;
  FCount := 0;
end;

procedure TStringAssociatedList.TrimToSize;
begin
  if FCount < FCapacity then
    SetCapacity(FCount);
end;

{ TStringHashTable }

const
  MissingStrPtr = Pointer($FFFFFFFF);

constructor TStringHashTable.Create(CaseSensitive: Boolean);
begin
  Create(CaseSensitive, 0);
end;

constructor TStringHashTable.Create(CaseSensitive: Boolean;
  InitialCapacity: Integer);
var
  N: Integer;
begin
  N := G_EnlargePrimeCapacity((InitialCapacity * 7) div 5);
  FCapacity := N;
  FMaxCount := (N * 5) div 7;
  FCaseSensitive := CaseSensitive;
  GetMem(FKeys, N shl 2);
  GetMem(FValues, N shl 2);
  G_FillLongs(0, FKeys, N);
end;

destructor TStringHashTable.Destroy;
var
  I: Integer;
  P: Pointer;
begin
  for I := FCapacity - 1 downto 0 do
  begin
    P := Pointer(FKeys^[I]);
    if (P <> nil) and (P <> MissingStrPtr) then
    begin
      FKeys^[I] := '';
      if FOwnValues then
        FValues^[I].Free;
    end;
  end;
  FreeMem(FKeys);
  FreeMem(FValues);
end;

procedure TStringHashTable.SetMaxCount(NewMaxCount: Integer);
var
  N: Integer;
begin
  if NewMaxCount >= FCount then
  begin
    N := G_EnlargePrimeCapacity((NewMaxCount * 7) div 5);
    if N <> FCapacity then
      SetCapacity(N);
  end;
end;

function TStringHashTable.GetIndex(P: Pointer): Integer;
var
  X, HashSize, HashStep: LongWord;
begin
  if not FCaseSensitive then
    X := G_CRC32_Text(P)
  else
    X := G_CRC32_Str(P);
  HashSize := LongWord(FCapacity);
  HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
  while True do
  begin
    X := X mod HashSize;
    if Pointer(FKeys^[X]) = nil then
    begin
      Result := X;
      Exit;
    end;
    Inc(X, HashStep);
  end;
end;

procedure TStringHashTable.SetCapacity(NewCapacity: Integer);
var
  PrevKeys: PStringItemList;
  PrevValues: PObjectItemList;
  P: Pointer;
  N, I: Integer;
begin
  N := NewCapacity;
  I := FCapacity - 1;
  PrevKeys := FKeys;
  PrevValues := FValues;
  FCapacity := N;
  FMaxCount := (N * 5) div 7;
  GetMem(FKeys, N shl 2);
  GetMem(FValues, N shl 2);
  G_FillLongs(0, FKeys, N);
  while I >= 0 do
  begin
    P := Pointer(PrevKeys^[I]);
    if (P <> nil) and (P <> MissingStrPtr) then
    begin
      N := GetIndex(P);
      Pointer(FKeys^[N]) := P;
      FValues^[N] := PrevValues^[I];
    end;
    Dec(I);
  end;
  FreeMem(PrevKeys);
  FreeMem(PrevValues);
end;

function TStringHashTable.GetItem(const Key: string): TObject;
var
  X, HashSize, HashStep: LongWord;
  P: Pointer;
begin
  HashSize := LongWord(FCapacity);
  if not FCaseSensitive then
  begin
    X := G_CRC32_Text(Key);
    HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
    while True do
    begin
      X := X mod HashSize;
      P := Pointer(FKeys^[X]);
      if P = nil then
        Break;
      if (P <> MissingStrPtr) and G_SameText(P, Pointer(Key)) then
      begin
        Result := FValues^[X];
        Exit;
      end;
      Inc(X, HashStep);
    end;
  end else
  begin
    X := G_CRC32_Str(Key);
    HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
    while True do
    begin
      X := X mod HashSize;
      P := Pointer(FKeys^[X]);
      if P = nil then
        Break;
      if (P <> MissingStrPtr) and G_SameStr(P, Pointer(Key)) then
      begin
        Result := FValues^[X];
        Exit;
      end;
      Inc(X, HashStep);
    end;
  end;
  Result := nil;
end;

procedure TStringHashTable.SetItem(const Key: string; Value: TObject);
var
  X, HashSize, HashStep: LongWord;
  L: Integer;
  P: Pointer;
  O: TObject;
begin
  HashSize := LongWord(FCapacity);
  L := -1;
  if not FCaseSensitive then
  begin
    X := G_CRC32_Text(Key);
    HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
    while True do
    begin
      X := X mod HashSize;
      P := Pointer(FKeys^[X]);
      if (P <> nil) and (P <> MissingStrPtr) then
      begin
        if G_SameText(P, Pointer(Key)) then
        begin
          if not FOwnValues then
            FValues^[X] := Value
          else
          begin
            O := FValues^[X];
            if O <> Value then
            begin
              FValues^[X] := Value;
              O.Free;
            end;
          end;
          Exit;
        end;
      end else
      begin
        if L < 0 then
          L := X;
        if P = nil then
          Break;
      end;
      Inc(X, HashStep);
    end;
  end else
  begin
    X := G_CRC32_Str(Key);
    HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
    while True do
    begin
      X := X mod HashSize;
      P := Pointer(FKeys^[X]);
      if (P <> nil) and (P <> MissingStrPtr) then
      begin
        if G_SameStr(P, Pointer(Key)) then
        begin
          if not FOwnValues then
            FValues^[X] := Value
          else
          begin
            O := FValues^[X];
            if O <> Value then
            begin
              FValues^[X] := Value;
              O.Free;
            end;
          end;
          Exit;
        end;
      end else
      begin
        if L < 0 then
          L := X;
        if P = nil then
          Break;
      end;
      Inc(X, HashStep);
    end;
  end;
  if FCount >= FMaxCount then
  begin
    SetCapacity(G_EnlargePrimeCapacity(FCapacity));
    L := GetIndex(Pointer(Key));
  end;
  Pointer(FKeys^[L]) := nil;
  FKeys^[L] := Key;
  FValues^[L] := Value;
  Inc(FCount);
end;

procedure TStringHashTable.EnsureCapacity(Capacity: Integer);
begin
  if FMaxCount < Capacity then
    SetCapacity(G_EnlargePrimeCapacity((Capacity * 7) div 5));
end;

procedure TStringHashTable.Add(const Key: string; Value: TObject);
var
  X, HashSize, HashStep: LongWord;
  L: Integer;
  P: Pointer;
begin
  HashSize := LongWord(FCapacity);
  L := -1;
  if not FCaseSensitive then
  begin
    X := G_CRC32_Text(Key);
    HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
    while True do
    begin
      X := X mod HashSize;
      P := Pointer(FKeys^[X]);
      if (P <> nil) and (P <> MissingStrPtr) then
      begin
        if G_SameText(P, Pointer(Key)) then
          RaiseErrorFmt(SErrKeyDuplicatesInAssociatedList, 'TStringHashTable');
      end else
      begin
        if L < 0 then
          L := X;
        if P = nil then
          Break;
      end;
      Inc(X, HashStep);
    end;
  end else
  begin
    X := G_CRC32_Str(Key);
    HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
    while True do
    begin
      X := X mod HashSize;
      P := Pointer(FKeys^[X]);
      if (P <> nil) and (P <> MissingStrPtr) then
      begin
        if G_SameStr(P, Pointer(Key)) then
          RaiseErrorFmt(SErrKeyDuplicatesInAssociatedList, 'TStringHashTable');
      end else
      begin
        if L < 0 then
          L := X;
        if P = nil then
          Break;
      end;
      Inc(X, HashStep);
    end;
  end;
  if FCount >= FMaxCount then
  begin
    SetCapacity(G_EnlargePrimeCapacity(FCapacity));
    L := GetIndex(Pointer(Key));
  end;
  Pointer(FKeys^[L]) := nil;
  FKeys^[L] := Key;
  FValues^[L] := Value;
  Inc(FCount);
end;

function TStringHashTable.Contains(const Key: string): Boolean;
var
  X, HashSize, HashStep: LongWord;
  P: Pointer;
begin
  HashSize := LongWord(FCapacity);
  if not FCaseSensitive then
  begin
    X := G_CRC32_Text(Key);
    HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
    while True do
    begin
      X := X mod HashSize;
      P := Pointer(FKeys^[X]);
      if P = nil then
        Break;
      if (P <> MissingStrPtr) and G_SameText(P, Pointer(Key)) then
      begin
        Result := True;
        Exit;
      end;
      Inc(X, HashStep);
    end;
  end else
  begin
    X := G_CRC32_Str(Key);
    HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
    while True do
    begin
      X := X mod HashSize;
      P := Pointer(FKeys^[X]);
      if P = nil then
        Break;
      if (P <> MissingStrPtr) and G_SameStr(P, Pointer(Key)) then
      begin
        Result := True;
        Exit;
      end;
      Inc(X, HashStep);
    end;
  end;
  Result := False;
end;

procedure TStringHashTable.Remove(const Key: string);
var
  X, HashSize, HashStep: LongWord;
  P: Pointer;
begin
  HashSize := LongWord(FCapacity);
  if not FCaseSensitive then
  begin
    X := G_CRC32_Text(Key);
    HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
    while True do
    begin
      X := X mod HashSize;
      P := Pointer(FKeys^[X]);
      if P = nil then
        Break;
      if (P <> MissingStrPtr) and G_SameText(P, Pointer(Key)) then
      begin
        FKeys^[X] := '';
        Pointer(FKeys^[X]) := MissingStrPtr;
        if FOwnValues then
          FValues^[X].Free;
        Dec(FCount);
        Exit;
      end;
      Inc(X, HashStep);
    end;
  end else
  begin
    X := G_CRC32_Str(Key);
    HashStep := (((X shr 5) + 1) mod (HashSize - 1)) + 1;
    while True do
    begin
      X := X mod HashSize;
      P := Pointer(FKeys^[X]);
      if P = nil then
        Break;
      if (P <> MissingStrPtr) and G_SameStr(P, Pointer(Key)) then
      begin
        FKeys^[X] := '';
        Pointer(FKeys^[X]) := MissingStrPtr;
        if FOwnValues then
          FValues^[X].Free;
        Dec(FCount);
        Exit;
      end;
      Inc(X, HashStep);
    end;
  end;
end;

procedure TStringHashTable.Clear;
var
  I: Integer;
  P: Pointer;
begin
  for I := FCapacity - 1 downto 0 do
  begin
    P := Pointer(FKeys^[I]);
    if (P <> nil) and (P <> MissingStrPtr) then
    begin
      FKeys^[I] := '';
      if FOwnValues then
        FValues^[I].Free;
    end;
  end;
  G_FillLongs(0, FKeys, FCapacity);
  FCount := 0;
end;

{ TLinkedList }

destructor TLinkedList.Destroy;
begin
  Clear;
end;

function TLinkedList.AddHead(Value: TObject): PLinkedListNode;
var
  P: PLinkedListNode;
begin
  GetMem(P, SizeOf(TLinkedListNode));
  P^.Value := Value;
  P^.OwnerList := Self;
  P^.PrevNode := nil;
  P^.NextNode := FHeadNode;
  if FHeadNode <> nil then
    FHeadNode^.PrevNode := P
  else
    FTailNode := P;
  FHeadNode := P;
  Result := P;
end;

function TLinkedList.AddTail(Value: TObject): PLinkedListNode;
var
  P: PLinkedListNode;
begin
  GetMem(P, SizeOf(TLinkedListNode));
  P^.Value := Value;
  P^.OwnerList := Self;
  P^.PrevNode := FTailNode;
  P^.NextNode := nil;
  if FTailNode <> nil then
    FTailNode^.NextNode := P
  else
    FHeadNode := P;
  FTailNode := P;
  Result := P;
end;

function TLinkedList.InsertBefore(Node: PLinkedListNode; Value: TObject): PLinkedListNode;
var
  P: PLinkedListNode;
begin
  if (Node = nil) or (Node^.OwnerList <> Self) then
    RaiseError(SErrNodeBelongsToAnotherList);
  GetMem(P, SizeOf(TLinkedListNode));
  P^.Value := Value;
  P^.OwnerList := Self;
  P^.PrevNode := Node^.PrevNode;
  P^.NextNode := Node;
  Node^.PrevNode := P;
  if P^.PrevNode <> nil then
    P^.PrevNode^.NextNode := P
  else
    FHeadNode := P;
  Result := P;
end;

function TLinkedList.InsertAfter(Node: PLinkedListNode; Value: TObject): PLinkedListNode;
var
  P: PLinkedListNode;
begin
  if (Node = nil) or (Node^.OwnerList <> Self) then
    RaiseError(SErrNodeBelongsToAnotherList);
  GetMem(P, SizeOf(TLinkedListNode));
  P^.Value := Value;
  P^.OwnerList := Self;
  P^.PrevNode := Node;
  P^.NextNode := Node^.NextNode;
  Node^.NextNode := P;
  if P^.NextNode <> nil then
    P^.NextNode^.PrevNode := P
  else
    FTailNode := P;
  Result := P;
end;

function TLinkedList.IsEmpty: Boolean;
begin
  Result := FHeadNode = nil;
end;

function TLinkedList.PopHeadValue: TObject;
var
  P: PLinkedListNode;
begin
  P := FHeadNode;
  if P <> nil then
  begin
    FHeadNode := P^.NextNode;
    if FHeadNode <> nil then
      FHeadNode^.PrevNode := nil
    else
      FTailNode := nil;
    Result := P^.Value;
    FreeMem(P);
  end else
    Result := nil;
end;

function TLinkedList.PopTailValue: TObject;
var
  P: PLinkedListNode;
begin
  P := FTailNode;
  if P <> nil then
  begin
    FTailNode := P^.PrevNode;
    if FTailNode <> nil then
      FTailNode^.NextNode := nil
    else
      FHeadNode := nil;
    Result := P^.Value;
    FreeMem(P);
  end else
    Result := nil;
end;

procedure TLinkedList.RemoveHead;
var
  P: PLinkedListNode;
begin
  P := FHeadNode;
  if P <> nil then
  begin
    FHeadNode := P^.NextNode;
    if FHeadNode <> nil then
      FHeadNode^.PrevNode := nil
    else
      FTailNode := nil;
    if FOwnValues then
      P^.Value.Free;
    FreeMem(P);
  end;
end;

procedure TLinkedList.RemoveTail;
var
  P: PLinkedListNode;
begin
  P := FTailNode;
  if P <> nil then
  begin
    FTailNode := P^.PrevNode;
    if FTailNode <> nil then
      FTailNode^.NextNode := nil
    else
      FHeadNode := nil;
    if FOwnValues then
      P^.Value.Free;
    FreeMem(P);
  end;
end;

procedure TLinkedList.Remove(Node: PLinkedListNode);
begin
  if (Node = nil) or (Node^.OwnerList <> Self) then
    RaiseError(SErrNodeBelongsToAnotherList);
  if Node = FHeadNode then
    RemoveHead
  else if Node = FTailNode then
    RemoveTail
  else
  begin
    Node^.PrevNode^.NextNode := Node^.NextNode;
    Node^.NextNode^.PrevNode := Node^.PrevNode;
    if FOwnValues then
      Node^.Value.Free;
    FreeMem(Node);
  end;
end;

procedure TLinkedList.Clear;
var
  P, N: PLinkedListNode;
begin
  P := FHeadNode;
  while P <> nil do
  begin
    N := P.NextNode;
    if FOwnValues then
      P.Value.Free;
    FreeMem(P);
    P := N;
  end;
  FHeadNode := nil;
  FTailNode := nil;
end;

{ TArrayList }

constructor TArrayList.Create;
begin
  Create(0);
end;

constructor TArrayList.Create(InitialCapacity: Integer);
begin
  FCapacity := InitialCapacity;
  if InitialCapacity > 0 then
    GetMem(FItems, InitialCapacity * SizeOf(Pointer));
end;

destructor TArrayList.Destroy;
begin
  if FCapacity > 0 then
    FreeMem(FItems);
end;

procedure TArrayList.Load(ArrayList: TArrayList);
var
  C: Integer;
begin
  C := ArrayList.FCount;
  if C > FCapacity then
  begin
    FCount := 0;
    SetCapacity(G_NormalizeCapacity(C));
  end;
  if C > 0 then
    G_CopyLongs(ArrayList.FItems, FItems, C);
  FCount := C;
end;

procedure TArrayList.Load(Items: Pointer; Count: Integer);
begin
  if Count > FCapacity then
  begin
    FCount := 0;
    SetCapacity(G_NormalizeCapacity(Count));
  end;
  if Count > 0 then
    G_CopyLongs(Items, FItems, Count);
  FCount := Count;
end;

procedure TArrayList.SetCapacity(NewCapacity: Integer);
var
  NewItems: PPointerItemList;
begin
  if (NewCapacity <> FCapacity) and (NewCapacity >= FCount) then
  begin
    if NewCapacity > 0 then
    begin
      GetMem(NewItems, NewCapacity * SizeOf(Pointer));
      if FCount > 0 then
        G_CopyLongs(FItems, NewItems, FCount);
    end else
      NewItems := nil;
    if FCapacity > 0 then
      FreeMem(FItems);
    FCapacity := NewCapacity;
    FItems := NewItems;
  end;
end;

procedure TArrayList.EnsureCapacity(Capacity: Integer);
begin
  if FCapacity < Capacity then
    SetCapacity(G_NormalizeCapacity(Capacity));
end;

procedure TArrayList.Add(P: Pointer);
begin
  if FCount >= FCapacity then
    SetCapacity(G_EnlargeCapacity(FCapacity));
  FItems^[FCount] := P;
  Inc(FCount);
end;

procedure TArrayList.Insert(Index: Integer; P: Pointer);
begin
  if FCount >= FCapacity then
    SetCapacity(G_EnlargeCapacity(FCapacity));
  if Index < FCount then
    G_MoveLongs(@FItems^[Index], @FItems^[Index + 1], FCount - Index);
  FItems^[Index] := P;
  Inc(FCount);
end;

function TArrayList.Peek: Pointer;
begin
  Result := FItems^[FCount - 1];
end;

function TArrayList.Pop: Pointer;
begin
  Dec(FCount);
  Result := FItems^[FCount];
end;

function TArrayList.IndexOf(P: Pointer): Integer;
begin
  Result := G_Scan_Pointer(P, FItems, FCount);
end;

procedure TArrayList.RemoveAt(Index: Integer);
begin
  Dec(FCount);
  if Index < FCount then
    G_MoveLongs(@FItems^[Index + 1], @FItems^[Index], FCount - Index);
end;

procedure TArrayList.RemoveAndFreeItemAt(Index: Integer);
begin
  TObject(FItems^[Index]).Free;
  RemoveAt(Index);
end;

procedure TArrayList.UnorderedRemoveAt(Index: Integer);
begin
  Dec(FCount);
  if Index < FCount then
    FItems^[Index] := FItems^[FCount];
end;

procedure TArrayList.UnorderedRemoveAndFreeItemAt(Index: Integer);
begin
  TObject(FItems^[Index]).Free;
  UnorderedRemoveAt(Index);
end;

procedure TArrayList.Clear;
begin
  FCount := 0;
end;

procedure TArrayList.ClearAndFreeItems;
var
  I: Integer;
begin
  for I := FCount - 1 downto 0 do
    TObject(FItems^[I]).Free;
  FCount := 0;
end;

procedure TArrayList.Sort(CompareFunction: TCompareFunction);
begin
  G_SortAscending(FItems, FCount, CompareFunction);
end;

function TArrayList.EnumerateGroups(CompareFunction: TCompareFunction): TGroupEnumerator;
begin
  Result := TGroupEnumerator.Create(FItems, FCount, CompareFunction);
end;

function TArrayList.Search(Value: Pointer; MatchFunction: TMatchFunction): Pointer;
var
  Index: Integer;
begin
  if Search(Value, MatchFunction, Index) then
    Result := FItems^[Index]
  else
    Result := nil;
end;

function TArrayList.Search(Value1, Value2: Pointer; MatchFunction: TMatch2Function): Pointer;
var
  Index: Integer;
begin
  if Search(Value1, Value2, MatchFunction, Index) then
    Result := FItems^[Index]
  else
    Result := nil;
end;

function TArrayList.Search(Value: Pointer; MatchFunction: TMatchFunction; out Index: Integer): Boolean;
var
  L, H, M: Integer;
begin
  L := 0;
  if FCount > 0 then
  begin
    H := FCount - 1;
    while L <= H do
    begin
      M := (L + H) shr 1;
      if MatchFunction(Value, FItems^[M]) > 0 then
        L := M + 1
      else
        H := M - 1;
    end;
    Result := (L < FCount) and (MatchFunction(Value, FItems^[L]) = 0);
  end else
    Result := False;
  Index := L;
end;

function TArrayList.Search(Value1, Value2: Pointer; MatchFunction: TMatch2Function; out Index: Integer): Boolean;
var
  L, H, M: Integer;
begin
  L := 0;
  if FCount > 0 then
  begin
    H := FCount - 1;
    while L <= H do
    begin
      M := (L + H) shr 1;
      if MatchFunction(Value1, Value2, FItems^[M]) > 0 then
        L := M + 1
      else
        H := M - 1;
    end;
    Result := (L < FCount) and (MatchFunction(Value1, Value2, FItems^[L]) = 0);
  end else
    Result := False;
  Index := L;
end;

procedure TArrayList.TrimToSize;
begin
  if FCount < FCapacity then
    SetCapacity(FCount);
end;

function TArrayList.Equals(ArrayList: TArrayList): Boolean;
begin
  if ArrayList = Self then
    Result := True
  else if (ArrayList <> nil) and (ArrayList.FCount = FCount) then
    Result := G_SameLongs(ArrayList.FItems, FItems, FCount)
  else
    Result := False;
end;

function TArrayList.Clone: TArrayList;
begin
  Result := TArrayList.Create(FCount);
  if FCount > 0 then
    G_CopyLongs(FItems, Result.FItems, FCount);
  Result.FCount := FCount;
end;

{ TArrayReadOnlyList }

constructor TArrayReadOnlyList.Create(ItemList: PPointerItemList; Count: Integer);
begin
  FItems := ItemList;
  FCount := Count;
end;

function TArrayReadOnlyList.IndexOf(P: Pointer): Integer;
begin
  Result := G_Scan_Pointer(P, FItems, FCount);
end;

procedure TArrayReadOnlyList.Sort(CompareFunction: TCompareFunction);
begin
  G_SortAscending(FItems, FCount, CompareFunction);
end;

function TArrayReadOnlyList.EnumerateGroups(CompareFunction: TCompareFunction): TGroupEnumerator;
begin
  Result := TGroupEnumerator.Create(FItems, FCount, CompareFunction);
end;

function TArrayReadOnlyList.Search(Value: Pointer; MatchFunction: TMatchFunction): Pointer;
var
  Index: Integer;
begin
  if Search(Value, MatchFunction, Index) then
    Result := FItems^[Index]
  else
    Result := nil;
end;

function TArrayReadOnlyList.Search(Value1, Value2: Pointer; MatchFunction: TMatch2Function): Pointer;
var
  Index: Integer;
begin
  if Search(Value1, Value2, MatchFunction, Index) then
    Result := FItems^[Index]
  else
    Result := nil;
end;

function TArrayReadOnlyList.Search(Value: Pointer; MatchFunction: TMatchFunction; out Index: Integer): Boolean;
var
  L, H, M: Integer;
begin
  L := 0;
  if FCount > 0 then
  begin
    H := FCount - 1;
    while L <= H do
    begin
      M := (L + H) shr 1;
      if MatchFunction(Value, FItems^[M]) > 0 then
        L := M + 1
      else
        H := M - 1;
    end;
    Result := (L < FCount) and (MatchFunction(Value, FItems^[L]) = 0);
  end else
    Result := False;
  Index := L;
end;

function TArrayReadOnlyList.Search(Value1, Value2: Pointer; MatchFunction: TMatch2Function; out Index: Integer): Boolean;
var
  L, H, M: Integer;
begin
  L := 0;
  if FCount > 0 then
  begin
    H := FCount - 1;
    while L <= H do
    begin
      M := (L + H) shr 1;
      if MatchFunction(Value1, Value2, FItems^[M]) > 0 then
        L := M + 1
      else
        H := M - 1;
    end;
    Result := (L < FCount) and (MatchFunction(Value1, Value2, FItems^[L]) = 0);
  end else
    Result := False;
  Index := L;
end;

function TArrayReadOnlyList.Equals(ArrayList: TArrayList): Boolean;
begin
  if (ArrayList <> nil) and (ArrayList.FCount = FCount) then
    Result := G_SameLongs(ArrayList.FItems, FItems, FCount)
  else
    Result := False;
end;

function TArrayReadOnlyList.Equals(ArrayList: TArrayReadOnlyList): Boolean;
begin
  if ArrayList = Self then
    Result := True
  else if (ArrayList <> nil) and (ArrayList.FCount = FCount) then
    Result := G_SameLongs(ArrayList.FItems, FItems, FCount)
  else
    Result := False;
end;

function TArrayReadOnlyList.Clone: TArrayList;
begin
  Result := TArrayList.Create(FCount);
  if FCount > 0 then
    G_CopyLongs(FItems, Result.FItems, FCount);
  Result.FCount := FCount;
end;

{ TGroupEnumerator }

constructor TGroupEnumerator.Create(ItemList: PPointerItemList; Count: Integer; CompareFunction: TCompareFunction);
var
  I, N, P: Integer;
begin
  if Count > 0 then
  begin
    G_SortAscending(ItemList, Count, CompareFunction);
    N := 1;
    for I := 1 to Count - 1 do
      if CompareFunction(ItemList^[I - 1], ItemList^[I]) <> 0 then
        Inc(N);
    FGroupCount := N;
    GetMem(FGroups, N * SizeOf(Pointer));
    N := 0;
    P := 0;
    for I := 1 to Count - 1 do
      if CompareFunction(ItemList^[I - 1], ItemList^[I]) <> 0 then
      begin
        FGroups^[N] := TArrayReadOnlyList.Create(@ItemList^[P], I - P);
        Inc(N);
        P := I;
      end;
    FGroups^[N] := TArrayReadOnlyList.Create(@ItemList^[P], Count - P);
  end;
end;

destructor TGroupEnumerator.Destroy;
var
  I: Integer;
begin
  if FGroupCount > 0 then
  begin
    for I := FGroupCount - 1 downto 0 do
      FGroups^[I].Free;
    FreeMem(FGroups);
  end;
end;

{   }

var
  ItemListToSort: PPointerItemList;
  CompareFunctionToSort: TCompareFunction;
  SortingCriticalSection: _RTL_CRITICAL_SECTION;

procedure QuickSortAsc(L, R: Integer);
var
  I, J: Integer;
  P, T: Pointer;
begin
  repeat
    I := L;
    J := R;
    P := ItemListToSort^[(L + R) shr 1];
    repeat
      while CompareFunctionToSort(ItemListToSort^[I], P) < 0 do Inc(I);
      while CompareFunctionToSort(ItemListToSort^[J], P) > 0 do Dec(J);
      if I <= J then
      begin
        T := ItemListToSort^[I];
        ItemListToSort^[I] := ItemListToSort^[J];
        ItemListToSort^[J] := T;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSortAsc(L, J);
    L := I;
  until I >= R;
end;

procedure G_SortAscending(ItemList: PPointerItemList; Count: Integer; CompareFunction: TCompareFunction);
begin
  if Count > 1 then
    try
      if IsMultiThread then
        EnterCriticalSection(SortingCriticalSection);
      ItemListToSort := ItemList;
      CompareFunctionToSort := CompareFunction;
      QuickSortAsc(0, Count - 1);
    finally
      if IsMultiThread then
        LeaveCriticalSection(SortingCriticalSection);
    end;
end;

procedure QuickSortDesc(L, R: Integer);
var
  I, J: Integer;
  P, T: Pointer;
begin
  repeat
    I := L;
    J := R;
    P := ItemListToSort^[(L + R) shr 1];
    repeat
      while CompareFunctionToSort(ItemListToSort^[I], P) > 0 do Inc(I);
      while CompareFunctionToSort(ItemListToSort^[J], P) < 0 do Dec(J);
      if I <= J then
      begin
        T := ItemListToSort^[I];
        ItemListToSort^[I] := ItemListToSort^[J];
        ItemListToSort^[J] := T;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSortAsc(L, J);
    L := I;
  until I >= R;
end;

procedure G_SortDescending(ItemList: PPointerItemList; Count: Integer; CompareFunction: TCompareFunction);
begin
  if Count > 1 then
    try
      if IsMultiThread then
        EnterCriticalSection(SortingCriticalSection);
      ItemListToSort := ItemList;
      CompareFunctionToSort := CompareFunction;
      QuickSortDesc(0, Count - 1);
    finally
      if IsMultiThread then
        LeaveCriticalSection(SortingCriticalSection);
    end;
end;

initialization
  InitializeCriticalSection(SortingCriticalSection);

finalization
  DeleteCriticalSection(SortingCriticalSection);

end.

